from msldap import logger
from msldap.authentication.ntlm.native import NTLMAUTHHandler, NTLMHandlerSettings
from wsnet.pyodide.clientauth import WSNETAuth
import enum

class ISC_REQ(enum.IntFlag):
	DELEGATE = 1
	MUTUAL_AUTH = 2
	REPLAY_DETECT = 4
	SEQUENCE_DETECT = 8
	CONFIDENTIALITY = 16
	USE_SESSION_KEY = 32
	PROMPT_FOR_CREDS = 64
	USE_SUPPLIED_CREDS = 128
	ALLOCATE_MEMORY = 256
	USE_DCE_STYLE = 512
	DATAGRAM = 1024
	CONNECTION = 2048
	CALL_LEVEL = 4096
	FRAGMENT_SUPPLIED = 8192
	EXTENDED_ERROR = 16384
	STREAM = 32768
	INTEGRITY = 65536
	IDENTIFY = 131072
	NULL_SESSION = 262144
	MANUAL_CRED_VALIDATION = 524288
	RESERVED1 = 1048576
	FRAGMENT_TO_FIT = 2097152
	HTTP = 0x10000000

class MSLDAPWSNetNTLMAuth:
	def __init__(self, settings):
		self.settings = settings
		self.mode = None #'CLIENT'
		self.sspi = WSNETAuth()
		self.operator = None
		self.client = None
		self.target = None
		self.seq_number = 0
		
		self.session_key = None
		self.ntlm_ctx = NTLMAUTHHandler(NTLMHandlerSettings(None, 'MANUAL'))

	def setup(self):
		return
		
	@property
	def ntlmChallenge(self):
		return self.ntlm_ctx.ntlmChallenge
		
	def get_sealkey(self, mode = 'Client'):
		return self.ntlm_ctx.get_sealkey(mode = mode)
			
	def get_signkey(self, mode = 'Client'):
		return self.ntlm_ctx.get_signkey(mode = mode)

	def signing_needed(self):
		return self.ntlm_ctx.signing_needed()
	
	def encryption_needed(self):
		return self.ntlm_ctx.encryption_needed()
		
	async def encrypt(self, data, message_no):
		return await self.ntlm_ctx.encrypt(data, message_no)
		
	async def decrypt(self, data, sequence_no, direction='init', auth_data=None):
		return await self.ntlm_ctx.decrypt(data, sequence_no, direction=direction, auth_data=auth_data)

	async def sign(self, data, message_no, direction=None, reset_cipher = False):
		return await self.ntlm_ctx.sign(data, message_no, direction=None, reset_cipher = reset_cipher)
		
		
	def SEAL(self, signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt):
		return self.ntlm_ctx.SEAL(signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt)
		
	def SIGN(self, signingKey, message, seqNum, cipher_encrypt):
		return self.ntlm_ctx.SIGN(signingKey, message, seqNum, cipher_encrypt)
	
	def get_session_key(self):
		return self.session_key

	def get_seq_number(self):
		return self.seq_number
		
	def is_extended_security(self):
		return self.ntlm_ctx.is_extended_security()
	
	async def authenticate(self, authData = b'', flags = None, seq_number = 0, cb_data = None):
		try:
			if flags is None:
				flags = ISC_REQ.CONNECTION
			
			if authData is None:
				status, ctxattr, data, err = await self.sspi.authenticate('NTLM', '', '', 3, flags.value, authdata = b'')
				if err is not None:
					raise err
				self.ntlm_ctx.load_negotiate(data)
				return data, True, None
			else:
				self.ntlm_ctx.load_challenge(authData)
				status, ctxattr, data, err = await self.sspi.authenticate('NTLM', '', '', 3, flags.value, authdata = authData)
				if err is not None:
					raise err
				if err is None:
					self.ntlm_ctx.load_authenticate(data)
					self.session_key, err = await self.sspi.get_sessionkey()
					if err is not None:
						raise err
					self.ntlm_ctx.load_sessionkey(self.get_session_key())
				
				await self.sspi.disconnect()
				return data, False, None
		except Exception as e:
			return None, None, e
		
	


import enum
import asyncio
import ipaddress
import copy

from asysocks.common.clienturl import SocksClientURL 
from asysocks.common.constants import SocksServerVersion, SocksProtocol, SOCKS5Method
from asysocks.common.target import SocksTarget

from msldap import logger
from msldap.network.socks import SocksProxyConnection
from msldap.commons.proxy import MSLDAPProxy, MSLDAPProxyType
from minikerberos.common.target import KerberosTarget
from minikerberos.common.proxy import KerberosProxy



class MultiplexorProxyConnection:
	"""
	"""
	def __init__(self, target):
		self.target = target
		
	async def connect(self, is_kerberos = False):
		"""
		
		"""		
		#hiding the import, so you'll only need to install multiplexor only when actually using it
		from multiplexor.operator import MultiplexorOperator
		
		con_str = self.target.proxy.target.get_server_url()
		#creating operator and connecting to multiplexor server
		self.operator = MultiplexorOperator(con_str, logging_sink = logger)
		await self.operator.connect()
		#creating socks5 proxy
		server_info = await self.operator.start_socks5(self.target.proxy.target.agent_id)
		await self.operator.terminate()
		#print(server_info)
		if is_kerberos is False:
			
			#copying the original target, then feeding it to socks5proxy object. it will hold the actual socks5 proxy server address we created before
			tp = MSLDAPProxy()
			tp.target = SocksTarget()
			tp.target.version = SocksServerVersion.SOCKS5
			tp.target.server_ip = server_info['listen_ip']
			tp.target.server_port = server_info['listen_port']
			tp.target.is_bind = False
			tp.target.proto = SocksProtocol.TCP
			tp.target.timeout = self.target.timeout
			tp.target.buffer_size = 4096
			
			tp.target.endpoint_ip = self.target.host
			tp.target.endpoint_port = self.target.port
			tp.target.endpoint_timeout = None # TODO: maybe implement endpoint timeout in the msldap target?
			tp.type = MSLDAPProxyType.SOCKS5

			newtarget = copy.deepcopy(self.target)
			newtarget.proxy = tp

			

			return SocksProxyConnection(target = newtarget)
		
		else:
			kt = copy.deepcopy(self.target)
			kt.proxy = KerberosProxy()
			kt.proxy.target = SocksTarget()
			kt.proxy.target.version = SocksServerVersion.SOCKS5
			kt.proxy.target.server_ip = server_info['listen_ip']
			kt.proxy.target.server_port = server_info['listen_port']
			kt.proxy.target.is_bind = False
			kt.proxy.target.proto = SocksProtocol.TCP
			kt.proxy.target.timeout = 10
			kt.proxy.target.buffer_size = 4096
			
			kt.proxy.target.endpoint_ip = self.target.ip
			kt.proxy.target.endpoint_port = self.target.port
			#kt.proxy.creds = copy.deepcopy(self.target.proxy.auth)
			
			return kt




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#


MSADSCHEMAENTRY_ATTRS = [ 	
	'cn', 'distinguishedName', 'adminDescription', 
	'adminDisplayName', 'objectGUID', 'schemaIDGUID', 
	'lDAPDisplayName', 'name', 
]

class MSADSchemaEntry:
	def __init__(self):
		self.cn = None #str
		self.distinguishedName = None #dn
		self.adminDescription = None #dunno
		self.adminDisplayName = None #datetime
		self.objectGUID = None #int
		self.schemaIDGUID = None
		self.lDAPDisplayName = None
		self.name = None #int

	
	@staticmethod
	def from_ldap(entry):
		adi = MSADSchemaEntry()
		adi.cn = entry['attributes'].get('cn') 
		adi.distinguishedName = entry['attributes'].get('distinguishedName')
		adi.adminDescription = entry['attributes'].get('adminDescription')
		adi.adminDisplayName = entry['attributes'].get('adminDisplayName')
		adi.objectGUID = entry['attributes'].get('objectGUID') #str
		adi.schemaIDGUID = entry['attributes'].get('schemaIDGUID') #list
		adi.lDAPDisplayName = entry['attributes'].get('lDAPDisplayName') #int
		adi.name = entry['attributes'].get('name') #int
		return adi
	
	def to_dict(self):
		d = {}
		d['cn'] = self.cn
		d['distinguishedName'] = self.distinguishedName
		d['adminDescription'] = self.adminDescription
		d['adminDisplayName'] = self.adminDisplayName
		d['objectGUID'] = self.objectGUID
		d['schemaIDGUID'] = self.schemaIDGUID
		d['lDAPDisplayName'] = self.lDAPDisplayName
		d['name'] = self.name
		return d


	def __str__(self):
		t = 'MSADSchemaEntry\r\n'
		d = self.to_dict()
		for k in d:
			t += '%s: %s\r\n' % (k, d[k])
		return t 


import copy
import asyncio

from msldap.commons.proxy import MSLDAPProxyType


class Proxyhandler:
	def __init__(self, target):
		self.target = target
	
	def select(self):
		if self.target.proxy is None:
			return self.target
		
		if self.target.proxy.proxy_type in [MSLDAPProxyType.SOCKS5, MSLDAPProxyType.SOCKS5_SSL]:
			import socket
			try:
				from socks5line.socks5line import Socks5LineProxyServer,SOCKS5Line 
			except ImportError:
				raise Exception('Failed to import socks5line proxy emulator! Install it then retry!')
			s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			s.bind(('127.0.0.1', 0))
			new_port = s.getsockname()[1]
			proxy = Socks5LineProxyServer()
			proxy.ip = self.target.proxy.ip
			proxy.port = self.target.proxy.port
			proxy.timeout = self.target.proxy.timeout
			proxy.username = self.target.proxy.username
			proxy.password = self.target.proxy.secret

			sl = SOCKS5Line(proxy, self.target.host, self.target.port)
			sl.run_newthread(s)

			newtarget = copy.deepcopy(self.target)
			newtarget.proxy = None
			newtarget.host = '127.0.0.1'
			newtarget.port = new_port

			return newtarget

		elif self.target.proxy.proxy_type in [LDAPProxyType.MULTIPLEXOR, LDAPProxyType.MULTIPLEXOR_SSL]:
			import socket
			try:
				from socks5line.socks5line import Socks5LineProxyServer,SOCKS5Line 
			except ImportError:
				raise Exception('Failed to import socks5line proxy emulator! Install it then retry!')
			try:
				from multiplexor.operator import MultiplexorOperator
			except ImportError:
				raise Exception('Failed to import multiplexor! Install it then retry!')

			async def create_proxy(connection_string, agent_id):
				try:
					#creating operator and connecting to multiplexor server
					self.operator = MultiplexorOperator(con_str, reconnect_tries = 1)
					await self.operator.connect()
					#creating socks5 proxy
					server_info = await self.operator.start_socks5(agent_id)
					asyncio.create_task(self.operator.terminate())
					return server_info
				except Exception as e:
					asyncio.create_task(self.operator.terminate())
					return e

			#creating connection string
			if self.target.proxy.proxy_type == LDAPProxyType.MULTIPLEXOR:
				con_str = 'ws://%s:%s' % (self.target.proxy.ip, self.target.proxy.port)
			else:
				con_str = 'wss://%s:%s' % (self.target.proxy.ip, self.target.proxy.port)
			

			#because of URL stuff, this logic needs to be in place
			#if self.target.proxy.domain is None:
			#	agent_id = self.target.proxy.username
			#else:
			#	agent_id = self.target.proxy.domain
			print('proxy_connecting')
			server_info = asyncio.run(create_proxy(con_str, self.target.proxy.settings['agentid'][0]))
			print('socks5 server info %s' % server_info)
			if isinstance(server_info, Exception):
				raise Exception('Failed to create socks proxy Reason: %s '% server_info)
			#copying the original target, then feeding it to socks5proxy object. it will hold the actual socks5 proxy server address we created before
			
			s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			s.bind(('127.0.0.1', 0))
			new_port = s.getsockname()[1]
			proxy = Socks5LineProxyServer()
			proxy.ip = server_info['listen_ip']
			proxy.port = server_info['listen_port']
			proxy.timeout = self.target.proxy.timeout
			proxy.username = self.target.proxy.username
			proxy.password = self.target.proxy.secret

			sl = SOCKS5Line(proxy, self.target.host, self.target.port)
			sl.run_newthread(s)
			
			print('socks5 socks5line ready')
			newtarget = copy.deepcopy(self.target)
			newtarget.proxy = None
			newtarget.host = '127.0.0.1'
			newtarget.port = new_port

			return newtarget


import io


from msldap.authentication.ntlm.structures.fields import Fields
from msldap.authentication.ntlm.structures.negotiate_flags import NegotiateFlags
from msldap.authentication.ntlm.structures.version import Version

# https://msdn.microsoft.com/en-us/library/cc236641.aspx
class NTLMNegotiate:
	def __init__(self):
		self.Signature         = b'NTLMSSP\x00'
		self.MessageType       = 1
		self.NegotiateFlags    = None
		self.DomainNameFields  = None
		self.WorkstationFields = None
		self.Version           = None
		self.Payload           = None

		####High-level variables
		self.Domain      = None
		self.Workstation = None

	@staticmethod
	def from_bytes(bbuff):
		return NTLMNegotiate.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = NTLMNegotiate()
		t.Signature         = buff.read(8)
		t.MessageType       = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		t.NegotiateFlags    = NegotiateFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		t.DomainNameFields  = Fields.from_buffer(buff)
		t.WorkstationFields = Fields.from_buffer(buff)

		if t.NegotiateFlags & NegotiateFlags.NEGOTIATE_VERSION: 
			t.Version = Version.from_buffer(buff)
			
			
		currPos = buff.tell()
		t.Payload = buff.read()

		#currPos = buff.tell()
		
		if t.DomainNameFields.length != 0:
			buff.seek(t.DomainNameFields.offset, io.SEEK_SET)
			raw_data = buff.read(t.WorkstationFields.length)
			#print(raw_data)
			#print(t.DomainNameFields.length)
			try:
				t.Domain = raw_data.decode('utf-16le')
			except UnicodeDecodeError:
				# yet another cool bug. 
				t.Domain = raw_data.decode('utf-8')

		if t.WorkstationFields.length != 0:
			buff.seek(t.WorkstationFields.offset, io.SEEK_SET)
			raw_data = buff.read(t.WorkstationFields.length)
			try:
				t.Workstation = raw_data.decode('utf-16le')
			except UnicodeDecodeError:
				# yet another cool bug. 
				t.Workstation = raw_data.decode('utf-8')

		#buff.seek(currPos, io.SEEK_SET)
		
		return t
		
	@staticmethod
	def construct(flags, domainname = None, workstationname = None, version = None):
		nego = NTLMNegotiate()
		nego.NegotiateFlags = flags
		nego.Payload = b''
		
		payload_pos = 32
		if flags & NegotiateFlags.NEGOTIATE_VERSION:
			if not version:
				raise Exception('NEGOTIATE_VERSION set but no Version supplied!')
			payload_pos += 8
			nego.Version = version
		
		
		
		if nego.NegotiateFlags & NegotiateFlags.NEGOTIATE_OEM_DOMAIN_SUPPLIED and domainname:
			data =  domainname.encode('utf-16le')
			nego.Payload += data
			nego.DomainNameFields  = Fields(len(data), payload_pos)
			payload_pos += len(data)
			nego.Domain      = data
		
		else:
			nego.DomainNameFields  = Fields(0,0)
		
		if nego.NegotiateFlags & NegotiateFlags.NEGOTIATE_OEM_WORKSTATION_SUPPLIED and workstationname:
			data =  workstationname.encode('utf-16le') 
			nego.Payload += data
			nego.WorkstationFields  = Fields(len(data), payload_pos)
			payload_pos += len(data)
			nego.Workstation = data
			
		else:
			nego.WorkstationFields  = Fields(0,0)
		
		return nego
		
	def to_bytes(self):
		t = b''
		t += self.Signature
		t += self.MessageType.to_bytes(4, byteorder = 'little', signed = False)
		t += self.NegotiateFlags.to_bytes(4, byteorder = 'little', signed = False)
		t += self.DomainNameFields.to_bytes()		
		t += self.WorkstationFields.to_bytes()
		if self.Version:
			t += self.Version.to_bytes()
		t += self.Payload
		return t

	def __repr__(self):
		t  = '== NTLMNegotiate ==\r\n'
		t += 'Signature  : %s\r\n' % repr(self.Signature)
		t += 'MessageType: %s\r\n' % repr(self.MessageType)
		t += 'NegotiateFlags: %s\r\n' % repr(self.NegotiateFlags)
		t += 'Version    : %s\r\n' % repr(self.Version)
		t += 'Domain     : %s\r\n' % repr(self.Domain)
		t += 'Workstation: %s\r\n' % repr(self.Workstation)
		
		return t
		
def test():
	test_reconstrut()
	test_construct()
	
def test_reconstrut(data = None):
	print('=== reconstruct===')
	if not data:
		nego_test_data = bytes.fromhex('4e544c4d5353500001000000978208e2000000000000000000000000000000000a00d73a0000000f')
	else:
		nego_test_data = data
	nego = NTLMNegotiate.from_bytes(nego_test_data)
	print(repr(nego))
	nego_test_data_verify = nego.to_bytes()
	assert nego_test_data == nego_test_data_verify
	
def test_construct():
	flags = NegotiateFlags.NEGOTIATE_56|NegotiateFlags.NEGOTIATE_KEY_EXCH|NegotiateFlags.NEGOTIATE_128|\
			NegotiateFlags.NEGOTIATE_VERSION|\
			NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|\
			NegotiateFlags.NEGOTIATE_ALWAYS_SIGN|NegotiateFlags.NEGOTIATE_NTLM|NegotiateFlags.NEGOTIATE_LM_KEY|\
			NegotiateFlags.NEGOTIATE_SIGN|NegotiateFlags.REQUEST_TARGET|NegotiateFlags.NTLM_NEGOTIATE_OEM|NegotiateFlags.NEGOTIATE_UNICODE|\
			NegotiateFlags.NEGOTIATE_OEM_WORKSTATION_SUPPLIED|NegotiateFlags.NEGOTIATE_OEM_DOMAIN_SUPPLIED
	nego = NTLMNegotiate.construct(flags, domainname = "alma.com", workstationname = "testjoe", version = Version.construct())
	nego.to_bytes()
	print(repr(nego))
	
	test_reconstrut(nego.to_bytes())
	
	flags = NegotiateFlags.NEGOTIATE_56|NegotiateFlags.NEGOTIATE_KEY_EXCH|NegotiateFlags.NEGOTIATE_128|\
			NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY|\
			NegotiateFlags.NEGOTIATE_ALWAYS_SIGN|NegotiateFlags.NEGOTIATE_NTLM|NegotiateFlags.NEGOTIATE_LM_KEY|\
			NegotiateFlags.NEGOTIATE_SIGN|NegotiateFlags.REQUEST_TARGET|NegotiateFlags.NTLM_NEGOTIATE_OEM|NegotiateFlags.NEGOTIATE_UNICODE|\
			NegotiateFlags.NEGOTIATE_OEM_WORKSTATION_SUPPLIED|NegotiateFlags.NEGOTIATE_OEM_DOMAIN_SUPPLIED
	nego = NTLMNegotiate.construct(flags, domainname = "alma.com", workstationname = "testjoe2")
	print(nego.to_bytes())
	print(repr(nego))
	
	test_reconstrut(nego.to_bytes())
	
if __name__ == '__main__':
	test()



import datetime

def timestamp2datetime(dt):
	"""
	Converting Windows timestamps to datetime.datetime format
	:param dt: Windows timestamp as array of bytes
	:type dt: bytearray
	:return: datetime.datetime
	"""
	us = int.from_bytes(dt, byteorder='little')/ 10.
	return datetime.datetime(1601, 1, 1) + datetime.timedelta(microseconds=us)
	
	
def datetime2timestamp(dt):
	delta = dt - datetime.datetime(1601, 1, 1)
	ns = int((delta / datetime.timedelta(microseconds=1)) * 10)
	return ns.to_bytes(8, 'little', signed = False)

def wrap(s, w):
	return [s[i:i + w] for i in range(0, len(s), w)]

def print_cert(cert, offset=0):
	cert = cert['tbs_certificate']
	blanks = " " * offset
	msg = [
			"Cert Subject: %s" % cert['subject']['common_name'],
			"Cert Serial: %s" % cert['serial_number'],
			"Cert Start: %s" % cert['validity']['not_before'],
			"Cert End: %s" % cert['validity']['not_after'],
			"Cert Issuer: %s" % cert['issuer']['common_name'],
		]
	return "{}{}".format(blanks, "\n{}".format(blanks).join(msg))

KNOWN_SIDS = {
    "S-1-0": "Null Authority",
    "S-1-0-0": "Nobody",
    "S-1-1": "World Authority",
    "S-1-1-0": "Everyone",
    "S-1-2": "Local Authority",
    "S-1-2-0": "Local",
    "S-1-3": "Creator Authority",
    "S-1-3-0": "Creator Owner",
    "S-1-3-1": "Creator Group",
    "S-1-3-4": "Owner Rights",
    "S-1-4": "Non-unique Authority",
    "S-1-5": "NT Authority",
    "S-1-5-1": "Dialup",
    "S-1-5-2": "Network",
    "S-1-5-3": "Batch",
    "S-1-5-4": "Interactive",
    "S-1-5-5-X-Y": "Logon Session",
    "S-1-5-6": "Service",
    "S-1-5-7": "Anonymous",
    "S-1-5-9": "Enterprise Domain Controllers",
    "S-1-5-10": "Principal Self",
    "S-1-5-11": "Authenticated Users",
    "S-1-5-12": "Restricted Code",
    "S-1-5-13": "Terminal Server Users",
    "S-1-5-14": "Remote Interactive Logon",
    "S-1-5-17": "IUSR",
    "S-1-5-18": "Local System",
    "S-1-5-19": "NT Authority Local Service",
    "S-1-5-20": "NT Authority Network Service",
    "S-1-5-32-544": "Administrators",
    "S-1-5-32-545": "Users",
    "S-1-5-32-546": "Guests",
    "S-1-5-32-547": "Power Users",
    "S-1-5-32-548": "Account Operators",
    "S-1-5-32-549": "Server Operators",
    "S-1-5-32-550": "Print Operators",
    "S-1-5-32-551": "Backup Operators",
    "S-1-5-32-552": "Replicators",
    "S-1-5-32-582": "Storage Replica Administrators",
    "S-1-5-64-10": "NTLM Authentication",
    "S-1-5-64-14": "SChannel Authentication",
    "S-1-5-64-21": "Digest Authentication",
    "S-1-5-80": "NT Service",
}



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import asyncio
from os import terminal_size
import traceback
import logging
import csv
import shlex
import datetime
import copy
import typing

from msldap.external.aiocmd.aiocmd import aiocmd
from msldap.external.asciitree.asciitree import LeftAligned
from tqdm import tqdm

from msldap import logger
from asysocks import logger as sockslogger
from msldap.client import MSLDAPClient
from msldap.commons.url import MSLDAPURLDecoder
from msldap.ldap_objects import MSADUser, MSADMachine, MSADUser_TSV_ATTRS

from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR
from winacl.dtyp.ace import ACCESS_ALLOWED_OBJECT_ACE, ADS_ACCESS_MASK, AceFlags, ACE_OBJECT_PRESENCE
from winacl.dtyp.sid import SID
from winacl.dtyp.guid import GUID

from msldap.ldap_objects.adcertificatetemplate import MSADCertificateTemplate, EX_RIGHT_CERTIFICATE_ENROLLMENT, CertificateNameFlag
from msldap.wintypes.asn1.sdflagsrequest import SDFlagsRequest


class MSLDAPClientConsole(aiocmd.PromptToolkitCmd):
	def __init__(self, url = None):
		aiocmd.PromptToolkitCmd.__init__(self, ignore_sigint=False) #Setting this to false, since True doesnt work on windows...
		self.conn_url = None
		if url is not None:
			self.conn_url = MSLDAPURLDecoder(url)
		self.connection = None
		self.adinfo = None
		self.ldapinfo = None
		self.domain_name = None

	async def do_login(self, url = None):
		"""Performs connection and login"""
		try:			
			if self.conn_url is None and url is None:
				print('Not url was set, cant do logon')
			if url is not None:
				self.conn_url = MSLDAPURLDecoder(url)

			logger.debug(self.conn_url.get_credential())
			logger.debug(self.conn_url.get_target())
			
			
			self.connection = self.conn_url.get_client()
			_, err = await self.connection.connect()
			if err is not None:
				raise err
			
			return True
		except:
			traceback.print_exc()
			return False

	async def do_ldapinfo(self, show = True):
		"""Prints detailed LDAP connection info (DSA)"""
		try:
			if self.ldapinfo is None:
				self.ldapinfo = self.connection.get_server_info()
			if show is True:
				for k in self.ldapinfo:
					print('%s : %s' % (k, self.ldapinfo[k]))
			return True
		except:
			traceback.print_exc()
			return False

	async def do_adinfo(self, show = True):
		"""Prints detailed Active Driectory info"""
		try:
			if self.adinfo is None:
				self.adinfo = self.connection._ldapinfo
				self.domain_name = self.adinfo.distinguishedName.replace('DC','').replace('=','').replace(',','.')
			if show is True:
				print(self.adinfo)
			return True
		except:
			traceback.print_exc()
			return False

	async def do_spns(self):
		"""Fetches kerberoastable user accounts"""
		try:
			await self.do_ldapinfo(False)
			async for user, err in self.connection.get_all_service_users():
				if err is not None:
					raise err
				print(user.sAMAccountName)
			
			return True
		except:
			traceback.print_exc()
			return False

	async def do_asrep(self):
		"""Fetches ASREP-roastable user accounts"""
		try:
			await self.do_ldapinfo(False)
			async for user, err in self.connection.get_all_knoreq_users():
				if err is not None:
					raise err
				print(user.sAMAccountName)
			return True
		except:
			traceback.print_exc()
			return False

	async def do_computeraddr(self):
		"""Fetches all computer accounts"""
		try:
			await self.do_adinfo(False)
			#machine_filename = '%s_computers_%s.txt' % (self.domain_name, datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
		
			async for machine, err in self.connection.get_all_machines(attrs=['sAMAccountName', 'dNSHostName']):
				if err is not None:
					raise err
					
				dns = machine.dNSHostName
				if dns is None:
					dns = '%s.%s' % (machine.sAMAccountName[:-1], self.domain_name)

				print(str(dns))
			return True
		except:
			traceback.print_exc()
			return False

	async def do_dump(self):
		"""Fetches ALL user and machine accounts from the domain with a LOT of attributes"""
		try:
			await self.do_adinfo(False)
			await self.do_ldapinfo(False)
			
			users_filename = 'users_%s.tsv' % datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
			pbar = tqdm(desc = 'Writing users to file %s' % users_filename)
			with open(users_filename, 'w', newline='', encoding = 'utf8') as f:
				async for user, err in self.connection.get_all_users():
					if err is not None:
						raise err
					pbar.update()
					f.write('\t'.join(user.get_row(MSADUser_TSV_ATTRS)))
			print('Users dump was written to %s' % users_filename)
			
			users_filename = 'computers_%s.tsv' % datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
			pbar = tqdm(desc = 'Writing computers to file %s' % users_filename)
			with open(users_filename, 'w', newline='', encoding = 'utf8') as f:
				async for user, err in self.connection.get_all_machines():
					if err is not None:
						raise err
					pbar.update()
					f.write('\t'.join(user.get_row(MSADUser_TSV_ATTRS)))
			print('Computer dump was written to %s' % users_filename)
			return True
		except:
			traceback.print_exc()
			return False

	async def do_query(self, query, attributes = None):
		"""Performs a raw LDAP query against the server. Secondary parameter is the requested attributes SEPARATED WITH COMMA (,)"""
		try:
			await self.do_ldapinfo(False)
			if attributes is None:
				attributes = '*'
			if attributes.find(','):
				attributes = attributes.split(',')
			logging.debug('Query: %s' % (query))
			logging.debug('Attributes: %s' % (attributes))
			async for entry, err in self.connection.pagedsearch(query, attributes):
				if err is not None:
					raise err
				print(entry)
			return True
		except:
			traceback.print_exc()
			return False

	async def do_tree(self, dn = None, level = 1):
		"""Prints a tree from the given DN (if not set, the top) and with a given depth (default: 1)"""
		try:
			await self.do_ldapinfo(False)
			if level is None:
				level = 1
			level = int(level)
			if dn is not None:
				try:
					int(dn)
				except:
					pass
				else:
					level = int(dn)
					dn = None
					
			if dn is None:
				await self.do_ldapinfo(False)
				dn = self.connection._tree
			logging.debug('Tree on %s' % dn)
			tree_data = await self.connection.get_tree_plot(dn, level)
			tr = LeftAligned()
			print(tr(tree_data))

			return True
		except:
			traceback.print_exc()
			return False

	async def do_user(self, samaccountname, to_print=True):
		"""Feteches a user object based on the sAMAccountName of the user"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)
			user, err = await self.connection.get_user(samaccountname)
			if err is not None:
				raise err

			if to_print is True:
				if user is None:
					print('User not found!')
				else:
					print(user)
			
			return user
		except:
			traceback.print_exc()
			return False

	async def do_machine(self, samaccountname):
		"""Feteches a machine object based on the sAMAccountName of the machine"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)
			machine, err = await self.connection.get_machine(samaccountname)
			if err is not None:
				raise err
			if machine is None:
				print('machine not found!')
			else:
				print(machine)
				####TEST
				x = SECURITY_DESCRIPTOR.from_bytes(machine.allowedtoactonbehalfofotheridentity)
				print(x)
			
			return True
		except:
			traceback.print_exc()
			return False

	async def do_schemaentry(self, cn):
		"""Feteches a schema object entry object based on the DN of the object (must start with CN=)"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)
			schemaentry, err = await self.connection.get_schemaentry(cn)
			if err is not None:
				raise err
			
			print(str(schemaentry))
			return True
		except:
			traceback.print_exc()
			return False

	async def do_allschemaentry(self):
		"""Feteches all schema object entry objects"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)
			async for schemaentry, err in self.connection.get_all_schemaentry():
				if err is not None:
					raise err
				
				print(str(schemaentry))
			return True
		except:
			traceback.print_exc()
			return False

	#async def do_addallowedtoactonbehalfofotheridentity(self, target_name, add_computer_name):
	#	"""Adds a SID to the msDS-AllowedToActOnBehalfOfOtherIdentity protperty of target_dn"""
	#	try:
	#		await self.do_ldapinfo(False)
	#		await self.do_adinfo(False)
	#
	#		try:
	#			new_owner_sid = SID.from_string(sid)
	#		except:
	#			print('Incorrect SID!')
	#			return False, Exception('Incorrect SID')
	#
	#
	#		target_sd = None
	#		if target_attribute is None or target_attribute == '':
	#			target_attribute = 'nTSecurityDescriptor'
	#			res, err = await self.connection.get_objectacl_by_dn(target_dn)
	#			if err is not None:
	#				raise err
	#			target_sd = SECURITY_DESCRIPTOR.from_bytes(res)
	#		else:
	#			
	#			query = '(distinguishedName=%s)' % target_dn
	#			async for entry, err in self.connection.pagedsearch(query, [target_attribute]):
	#				if err is not None:
	#					raise err
	#				print(entry['attributes'][target_attribute])
	#				target_sd = SECURITY_DESCRIPTOR.from_bytes(entry['attributes'][target_attribute])
	#				break
	#			else:
	#				print('Target DN not found!')
	#				return False, Exception('Target DN not found!')
	#
	#		print(target_sd)
	#		new_sd = copy.deepcopy(target_sd)
	#		new_sd.Owner = new_owner_sid
	#		print(new_sd)
	#
	#		changes = {
	#			target_attribute : [('replace', [new_sd.to_bytes()])]
	#		}
	#		_, err = await self.connection.modify(target_dn, changes)
	#		if err is not None:
	#			raise err
	#
	#		print('Change OK!')
	#	except:
	#		traceback.print_exc()

	async def do_changeowner(self, new_owner_sid, target_dn, target_attribute = None):
		"""Changes the owner in a Security Descriptor to the new_owner_sid on an LDAP object or on an LDAP object's attribute identified by target_dn and target_attribute. target_attribute can be omitted to change the target_dn's SD's owner"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)

			_, err = await self.connection.change_priv_owner(new_owner_sid, target_dn, target_attribute = target_attribute)
			if err is not None:
				raise err
		except:
			traceback.print_exc()
			return False

	async def do_addprivdcsync(self, user_dn, forest = None):
		"""Adds DCSync rights to the given user by modifying the forest's Security Descriptor to add GetChanges and GetChangesAll ACE"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)

			_, err = await self.connection.add_priv_dcsync(user_dn, self.adinfo.distinguishedName)
			if err is not None:
				raise err

			print('Change OK!')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_addprivaddmember(self, user_dn, group_dn):
		"""Adds AddMember rights to the user on the group specified by group_dn"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)

			_, err = await self.connection.add_priv_addmember(user_dn, group_dn)
			if err is not None:
				raise err

			print('Change OK!')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_setsd(self, target_dn, sddl):
		"""Updates the security descriptor of an object"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)
	
			try:
				new_sd = SECURITY_DESCRIPTOR.from_sddl(sddl)
			except:
				print('Incorrect SDDL input!')
				return False, Exception('Incorrect SDDL input!')
	
			_, err = await self.connection.set_objectacl_by_dn(target_dn, new_sd.to_bytes())
			if err is not None:
				raise err
			print('Change OK!')
			return True
		except:
			print('Erro while updating security descriptor!')
			traceback.print_exc()
			return False
			
	async def do_getsd(self, dn):
		"""Feteches security info for a given DN"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)
			sec_info, err = await self.connection.get_objectacl_by_dn(dn)
			if err is not None:
				raise err
			sd = SECURITY_DESCRIPTOR.from_bytes(sec_info)
			print(sd.to_sddl())
			return True
		except:
			traceback.print_exc()
			return False

	async def do_gpos(self):
		"""Feteches security info for a given DN"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)
			async for gpo, err in self.connection.get_all_gpos():
				if err is not None:
					raise err
				print(gpo)
			
			return True
		except:
			traceback.print_exc()
			return False

	async def do_laps(self):
		"""Feteches all laps passwords"""
		try:
			async for entry, err in self.connection.get_all_laps():
				if err is not None:
					raise err
				pwd = '<MISSING>'
				if 'ms-Mcs-AdmPwd' in entry['attributes']:
					pwd = entry['attributes']['ms-Mcs-AdmPwd']
				print('%s : %s' % (entry['attributes']['cn'], pwd))
			
			return True
		except:
			traceback.print_exc()
			return False

	async def do_groupmembership(self, dn):
		"""Feteches names all groupnames the user is a member of for a given DN"""
		try:
			await self.do_ldapinfo(False)
			await self.do_adinfo(False)
			group_sids = []
			async for group_sid, err in self.connection.get_tokengroups(dn):
				if err is not None:
					raise err
				group_sids.append(group_sids)
				group_dn, err = await self.connection.get_dn_for_objectsid(group_sid)
				if err is not None:
					raise err
				print('%s - %s' % (group_dn, group_sid))
				
			if len(group_sids) == 0:
				print('No memberships found')
			
			return True
		except Exception as e:
			traceback.print_exc()
			return False

	async def do_bindtree(self, newtree):
		"""Changes the LDAP TREE for future queries. 
				 MUST be DN format eg. 'DC=test,DC=corp'
				 !DANGER! Switching tree to a tree outside of the domain will trigger a connection to that domain, leaking credentials!"""
		self.connection._tree = newtree
	
	async def do_trusts(self):
		"""Feteches gives back domain trusts"""
		try:
			async for entry, err in self.connection.get_all_trusts():
				if err is not None:
					raise err
				print(entry.get_line())
			
			return True
		except:
			traceback.print_exc()
			return False

	async def do_adduser(self, user_dn, password):
		"""Creates a new domain user with password"""
		try:
			_, err = await self.connection.create_user_dn(user_dn, password)
			if err is not None:
				raise err
			print('User added')
			return True
		except:
			traceback.print_exc()
			return False

	
	async def do_deluser(self, user_dn):
		"""Deletes the user! This action is irrecoverable (actually domain admins can do that but probably will shout with you)"""
		try:
			_, err = await self.connection.delete_user(user_dn)
			if err is not None:
				raise err
			print('Goodbye, Caroline.')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_changeuserpw(self, user_dn, newpass, oldpass = None):
		"""Changes user password, if you are admin then old pw doesnt need to be supplied"""
		try:
			_, err = await self.connection.change_password(user_dn, newpass, oldpass)
			if err is not None:
				raise err
			print('User password changed')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_unlockuser(self, user_dn):
		"""Unlock user by setting lockoutTime to 0"""
		try:
			_, err = await self.connection.unlock_user(user_dn)
			if err is not None:
				raise err
			print('User unlocked')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_enableuser(self, user_dn):
		"""Unlock user by flipping useraccountcontrol bits"""
		try:
			_, err = await self.connection.enable_user(user_dn)
			if err is not None:
				raise err
			print('User enabled')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_disableuser(self, user_dn):
		"""Unlock user by flipping useraccountcontrol bits"""
		try:
			_, err = await self.connection.disable_user(user_dn)
			if err is not None:
				raise err
			print('User disabled')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_addspn(self, user_dn, spn):
		"""Adds an SPN entry to the users account"""
		try:
			_, err = await self.connection.add_user_spn(user_dn, spn)
			if err is not None:
				raise err
			print('SPN added!')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_addhostname(self, user_dn, hostname):
		"""Adds additional hostname to computer account"""
		try:
			_, err = await self.connection.add_additional_hostname(user_dn, hostname)
			if err is not None:
				raise err
			print('Hostname added!')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_addusertogroup(self, user_dn, group_dn):
		"""Adds user to specified group. Both user and group must be in DN format!"""
		try:
			_, err = await self.connection.add_user_to_group(user_dn, group_dn)
			if err is not None:
				raise err
			print('User added to group!')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_deluserfromgroup(self, user_dn, group_dn):
		"""Removes user from specified group. Both user and group must be in DN format!"""
		try:
			_, err = await self.connection.del_user_from_group(user_dn, group_dn)
			if err is not None:
				raise err
			print('User added to group!')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_rootcas(self, to_print = True):
		"""Lists Root CA certificates"""
		try:
			cas = []
			async for ca, err in self.connection.list_root_cas():
				if err is not None:
					raise err
				cas.append(ca)
				if to_print is True:
					print(ca)
			return cas
		except:
			traceback.print_exc()
			return False

	async def do_ntcas(self, to_print = True):
		"""Lists NT CA certificates"""
		try:
			cas = []
			async for ca, err in self.connection.list_ntcas():
				if err is not None:
					raise err
				cas.append(ca)
				if to_print is True:
					print(ca)
			return cas
		except:
			traceback.print_exc()
			return False
	
	async def do_aiacas(self, to_print = True):
		"""Lists AIA CA certificates"""
		try:
			cas = []
			async for ca, err in self.connection.list_aiacas():
				if err is not None:
					raise err
				cas.append(ca)
				if to_print is True:
					print(ca)
			return cas
		except:
			traceback.print_exc()
			return False

	async def do_enrollmentservices(self, to_print=True):
		"""Lists AIA CA certificates"""
		try:
			services = []
			async for srv, err in self.connection.list_enrollment_services():
				if err is not None:
					raise err
				services.append(srv)
				if to_print is True:
					print(srv)
			return services
		except:
			traceback.print_exc()
			return False

	async def do_addcerttemplatenameflagaltname(self, certtemplatename, flags = None):
		"""Modifyies the msPKI-Certificate-Name-Flag value of the specified certificate template and enables ENROLLEE_SUPPLIES_SUBJECT_ALT_NAME bit. If 'flags' is present then it will assign that value."""
		try:
			template = None
			async for template, err in self.connection.list_certificate_templates(certtemplatename):
				if err is not None:
					raise err
				break
			
			if template is None:
				raise Exception("Template could not be found!")
			
			template = typing.cast(MSADCertificateTemplate, template)
			if flags is not None:
				flags = int(flags)
			else:
				flags = int(CertificateNameFlag(template.Certificate_Name_Flag) | CertificateNameFlag.ENROLLEE_SUPPLIES_SUBJECT_ALT_NAME)

			changes = {
				'msPKI-Certificate-Name-Flag' : [('replace', [flags])]
			}
	
			_, err = await self.connection.modify(template.distinguishedName, changes)
			if err is not None:
				raise err
			
			print('Modify OK!')
			return True


		except:
			traceback.print_exc()
			return False

	async def do_addenrollmentright(self, certtemplatename, user_dn):
		"""Grants enrollment rights to a user (by DN) for the specified certificate template."""
		try:
			user_sid, err = await self.connection.get_objectsid_for_dn(user_dn)
			if err is not None:
				raise err
			
			template = None
			async for template, err in self.connection.list_certificate_templates(certtemplatename):
				if err is not None:
					raise err
				break
			
			if template is None:
				raise Exception("Template could not be found!")
			template = typing.cast(MSADCertificateTemplate, template)
			new_sd = copy.deepcopy(template.nTSecurityDescriptor)
			ace = ACCESS_ALLOWED_OBJECT_ACE()
			ace.Sid = SID.from_string(user_sid)
			ace.ObjectType = GUID.from_string(EX_RIGHT_CERTIFICATE_ENROLLMENT)
			ace.AceFlags = AceFlags(0)
			ace.Mask = ADS_ACCESS_MASK.READ_PROP | ADS_ACCESS_MASK.WRITE_PROP | ADS_ACCESS_MASK.CONTROL_ACCESS
			ace.Flags = ACE_OBJECT_PRESENCE.ACE_OBJECT_TYPE_PRESENT
			new_sd.Dacl.aces.append(ace)
			_, err = await self.connection.set_objectacl_by_dn(template.distinguishedName, new_sd.to_bytes(), flags=SDFlagsRequest.DACL_SECURITY_INFORMATION)
			if err is not None:
				raise err
			print('SD set sucessfully')
			return True
		except:
			traceback.print_exc()
			return False

	async def do_certtemplates(self, name = None, to_print = True):
		"""Lists certificate templates"""
		try:
			services = await self.do_enrollmentservices(to_print=False)
			templates = []
			async for template, err in self.connection.list_certificate_templates(name):
				if err is not None:
					raise err
				
				lt = None
				if template.nTSecurityDescriptor is not None:
					lt, err = await self.connection.resolv_sd(template.nTSecurityDescriptor)
					if err is not None:
						raise err
				template.sid_lookup_table = lt
				for srv in services:
					if template.name in srv.certificateTemplates:
						template.enroll_services.append('%s\\%s' % (srv.dNSHostName, srv.name))

				templates.append(template)
				if to_print is True:
					print(template.prettyprint())

			return templates
		except:
			traceback.print_exc()
			return False

	async def do_sidresolv(self, sid, to_print = True):
		"""Returns the domain and username for SID"""
		try:
			domain, username, err = await self.connection.resolv_sid(sid)
			if err is not None:
				raise err
			res = '%s\\%s' % (domain, username)
			if to_print is True:
				print(res)
			return res
		except:
			traceback.print_exc()
			return False

	async def do_certify(self, cmd = None, username = None):
		"""ADCA security test"""
		try:
			es = await self.do_enrollmentservices(to_print=False)
			if es is False:
				raise Exception('Listing enrollment Services error! %s' % es)
			if es is None:
				raise Exception('No Enrollment Services present, stopping!')
			
			templates = await self.do_certtemplates(to_print=False)
			if templates is False:
				raise Exception('Listing templates error! %s' % es)
			
			if templates is None:
				raise Exception('No templates exists!')
			
			for enrollment in es:
				print(enrollment)
			
			if cmd is not None:
				if cmd.lower().startswith('vuln') is True:
					tokengroups = None
					if username is not None:
						tokengroups, err = await self.connection.get_tokengroups_user(username)
						if err is not None:
							raise err

					for template in templates:
						isvuln, reason = template.is_vulnerable(tokengroups)
						if isvuln is True:
							print(reason)
							print(template)
			else:
				for template in templates:
					print(template)

			return True
		except:
			traceback.print_exc()
			return False

	async def do_whoamiraw(self):
		"""Simple whoami"""
		try:
			res, err = await self.connection.whoami()
			if err is not None:
				raise err
			print(res)
		except:
			traceback.print_exc()
			return False
		
	async def do_whoami(self):
		"""Full whoami"""
		try:
			res, err = await self.connection.whoamifull()
			if err is not None:
				raise err
			
			for x in res:
				if isinstance(res[x], str) is True:
					print('%s: %s' % (x, res[x]))
				elif isinstance(res[x], dict) is True:
					for k in res[x]:
						print('Group: %s (%s)' % (k,'\\'.join(res[x][k])))
			return True
		except:
			traceback.print_exc()
			return False, None

	async def do_test(self):
		"""testing, dontuse"""
		try:
			async for entry, err in self.connection.get_all_objectacl():
				if err is not None:
					raise err

				if entry.objectClass[-1] != 'user':
					print(entry.objectClass)

			return True
		except:
			traceback.print_exc()
			return False

	"""
	async def do_info(self):
		try:

		except Exception as e:
			traceback.print_exc()
	"""


async def amain(args):
	client = MSLDAPClientConsole(args.url)

	if len(args.commands) == 0:
		if args.no_interactive is True:
			print('Not starting interactive!')
			return
		await client.run()
	else:
		for command in args.commands:
			if command == 'i':
				await client.run()
				return
			cmd = shlex.split(command)
			res = await client._run_single_command(cmd[0], cmd[1:])
			if res is False:
				return

def main():
	import argparse
	parser = argparse.ArgumentParser(description='MS LDAP library')
	parser.add_argument('-v', '--verbose', action='count', default=0, help='Verbosity, can be stacked')
	parser.add_argument('-n', '--no-interactive', action='store_true')
	parser.add_argument('url', help='Connection string in URL format.')
	parser.add_argument('commands', nargs='*', help="Takes a series of commands which will be executed until error encountered. If the command is 'i' is encountered during execution it drops back to interactive shell.")

	args = parser.parse_args()
	
	
	###### VERBOSITY
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	else:
		sockslogger.setLevel(logging.DEBUG)
		logger.setLevel(logging.DEBUG)
		logging.basicConfig(level=logging.DEBUG)

	asyncio.run(amain(args))

	

if __name__ == '__main__':
	main()




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import platform
import hashlib
import getpass
import base64
import enum
import copy
from urllib.parse import urlparse, parse_qs

from msldap.commons.credential import MSLDAPCredential, LDAPAuthProtocol, MSLDAP_KERBEROS_PROTOCOLS
from msldap.commons.target import MSLDAPTarget, LDAPProtocol
from msldap.commons.proxy import MSLDAPProxy, MSLDAPProxyType
from msldap.client import MSLDAPClient
from msldap.connection import MSLDAPClientConnection

class PLAINTEXTSCHEME(enum.Enum):
	"""
	Additional conveinence functions.
	"""
	SIMPLE_PROMPT = 'SIMPLE_PROMPT'
	SIMPLE_HEX = 'SIMPLE_HEX'
	SIMPLE_B64 = 'SIMPLE_B64'
	PLAIN_PROMPT = 'PLAIN_PROMPT'
	PLAIN_HEX = 'PLAIN_HEX'
	PLAIN_B64 = 'PLAIN_B64'
	SICILY_PROMPT = 'SICILY_PROMPT'
	SICILY_HEX = 'SICILY_HEX'
	SICILY_B64 = 'SICILY_B64'
	NTLM_PROMPT = 'NTLM_PROMPT'
	NTLM_HEX = 'NTLM_HEX'
	NTLM_B64 = 'NTLM_B64'

class MSLDAPURLDecoder:
	"""
	The URL describes both the connection target and the credentials. This class creates all necessary objects to set up the client.
	
	:param url: 
	:type url: str
	"""

	help_epilog = """
	MSLDAP URL Format: <protocol>+<auth>://<username>:<password>@<ip_or_host>:<port>/<tree>/?<param>=<value>
	<protocol> sets the ldap protocol following values supported:
		- ldap
		- ldaps
	<auth> can be omitted if plaintext authentication is to be performed (in that case it default to ntlm-password), otherwise:
		- ntlm-password
		- ntlm-nt
		- kerberos-password (dc option param must be used)
		- kerberos-rc4 / kerberos-nt (dc option param must be used)
		- kerberos-aes (dc option param must be used)
		- kerberos-keytab (dc option param must be used)
		- kerberos-ccache (dc option param must be used)
		- sspi-ntlm (windows only!)
		- sspi-kerberos (windows only!)
		- anonymous
		- plain
		- simple
		- sicily (same format as ntlm-nt but using the SICILY authentication)
	<tree>:
		OPTIONAL. Specifies the root tree of all queries
	<param> can be:
		- timeout : connction timeout in seconds
		- proxytype: currently only socks5 proxy is supported
		- proxyhost: Ip or hostname of the proxy server
		- proxyport: port of the proxy server
		- proxytimeout: timeout in secodns for the proxy connection
		- dc: the IP address of the domain controller, MUST be used for kerberos authentication
		- encrypt: enable encryption. Only for NTLM. DOESNT WORK WITH LDAPS
		- etype: Supported encryption types for Kerberos authentication. Multiple can be specified.
		- rate: LDAP paged search query rate limit. Will sleep for seconds between each new page. Default: 0 (no limit)
		- pagesize: LDAP paged search query size per page. Max: 1000. Default: 1000

	Examples:
	ldap://10.10.10.2 (anonymous bind)
	ldaps://test.corp (anonymous bind)
	ldap+sspi-ntlm://test.corp
	ldap+sspi-kerberos://test.corp
	ldap://TEST\\victim:<password>@10.10.10.2 (defaults to SASL GSSAPI NTLM)
	ldap+simple://TEST\\victim:<password>@10.10.10.2 (SASL SIMPLE auth)
	ldap+plain://TEST\\victim:<password>@10.10.10.2 (SASL SIMPLE auth)
	ldap+ntlm-password://TEST\\victim:<password>@10.10.10.2
	ldap+ntlm-nt://TEST\\victim:<nthash>@10.10.10.2
	ldap+kerberos-password://TEST\\victim:<password>@10.10.10.2
	ldap+kerberos-rc4://TEST\\victim:<rc4key>@10.10.10.2
	ldap+kerberos-aes://TEST\\victim:<aes>@10.10.10.2
	ldap://TEST\\victim:password@10.10.10.2/DC=test,DC=corp/
	ldap://TEST\\victim:password@10.10.10.2/DC=test,DC=corp/?timeout=99&proxytype=socks5&proxyhost=127.0.0.1&proxyport=1080&proxytimeout=44
	"""
	
	def __init__(self, url, credential:MSLDAPCredential = None, target:MSLDAPTarget = None ):
		self.credential = credential
		self.target = target
		self.url = url
		self.ldap_scheme = None
		self.auth_scheme = None

		self.domain = None
		self.username = None
		self.password = None
		self.encrypt = False
		self.auth_settings = {}
		self.etypes = None
		self.altname = None
		self.altdomain = None

		self.ldap_proto = None
		self.ldap_host = None
		self.ldap_port = 389
		self.ldap_tree = None
		self.target_timeout = 5
		self.target_pagesize = 1000
		self.target_ratelimit = 0
		self.dc_ip = None
		self.serverip = None
		self.proxy = None

		self.__pwpreprocess = None

		if url is not None:
			self.parse()


	def get_credential(self):
		"""
		Creates a credential object
		
		:return: Credential object
		:rtype: :class:`MSLDAPCredential`
		"""
		if self.credential is not None:
			return copy.deepcopy(self.credential)
		t = MSLDAPCredential(
			domain=self.domain, 
			username=self.username, 
			password = self.password, 
			auth_method=self.auth_scheme, 
			settings = self.auth_settings,
			altname=self.altname,
			altdomain=self.altdomain
		)
		t.encrypt = self.encrypt
		t.etypes = self.etypes
		
		return t

	def get_target(self):
		"""
		Creates a target object
		
		:return: Target object
		:rtype: :class:`MSLDAPTarget`
		"""
		if self.target is not None:
			return copy.deepcopy(self.target)

		target = MSLDAPTarget(
			self.ldap_host, 
			port = self.ldap_port, 
			proto = self.ldap_scheme, 
			tree=self.ldap_tree,
			timeout = self.target_timeout,
			ldap_query_page_size = self.target_pagesize,
			ldap_query_ratelimit = self.target_ratelimit
		)
		target.domain = self.domain
		target.dc_ip = self.dc_ip
		target.proxy = self.proxy
		target.serverip = self.serverip
		return target
	
	def __str__(self):
		t = '==== MSLDAPURLDecoder ====\r\n'
		for k in self.__dict__:
			val = self.__dict__[k]
			if isinstance(val, enum.IntFlag):
				val = val
			elif isinstance(val, enum.Enum):
				val = val.name
			
			t += '%s: %s\r\n' % (k, str(val))
			
		return t


	def get_client(self):
		"""
		Creates a client that can be used to interface with the server
		
		:return: LDAP client
		:rtype: :class:`MSLDAPClient`
		"""
		cred = self.get_credential()
		target = self.get_target()
		return MSLDAPClient(target, cred)
	
	def get_connection(self):
		"""
		Creates a connection that can be used to interface with the server
		
		:return: LDAP connection
		:rtype: :class:`MSLDAPClientConnection`
		"""
		cred = self.get_credential()
		target = self.get_target()
		return MSLDAPClientConnection(target, cred)

	def scheme_decoder(self, scheme):
		schemes = []
		for item in scheme.upper().split('+'):
			schemes.append(item.replace('-','_'))

		if schemes[0] == 'LDAP':
			self.ldap_scheme = LDAPProtocol.TCP
			self.ldap_port = 389
		elif schemes[0] == 'LDAPS':
			self.ldap_scheme = LDAPProtocol.SSL
			self.ldap_port = 636
		elif schemes[0] == 'LDAP_SSL':
			self.ldap_scheme = LDAPProtocol.SSL
			self.ldap_port = 636
		elif schemes[0] == 'LDAP_TCP':
			self.ldap_scheme = LDAPProtocol.TCP
			self.ldap_port = 389
		elif schemes[0] == 'LDAP_UDP':
			self.ldap_scheme = LDAPProtocol.UDP
			self.ldap_port = 389
		else:
			raise Exception('Unknown protocol! %s' % schemes[0])
		
		if len(schemes) == 1:
			return
		
		try:
			x = PLAINTEXTSCHEME(schemes[1])
			if x == PLAINTEXTSCHEME.SIMPLE_PROMPT:
				self.auth_scheme = LDAPAuthProtocol.SIMPLE
				self.__pwpreprocess = 'PROMPT'

			if x == PLAINTEXTSCHEME.SIMPLE_HEX:
				self.auth_scheme = LDAPAuthProtocol.SIMPLE
				self.__pwpreprocess = 'HEX'

			if x == PLAINTEXTSCHEME.SIMPLE_B64:
				self.auth_scheme = LDAPAuthProtocol.SIMPLE
				self.__pwpreprocess = 'B64'

			if x == PLAINTEXTSCHEME.PLAIN_PROMPT:
				self.auth_scheme = LDAPAuthProtocol.PLAIN
				self.__pwpreprocess = 'PROMPT'

			if x == PLAINTEXTSCHEME.PLAIN_HEX:
				self.auth_scheme = LDAPAuthProtocol.PLAIN
				self.__pwpreprocess = 'HEX'

			if x == PLAINTEXTSCHEME.PLAIN_B64:
				self.auth_scheme = LDAPAuthProtocol.PLAIN
				self.__pwpreprocess = 'B64'

			if x == PLAINTEXTSCHEME.SICILY_PROMPT:
				self.auth_scheme = LDAPAuthProtocol.SICILY
				self.__pwpreprocess = 'PROMPT'

			if x == PLAINTEXTSCHEME.SICILY_HEX:
				self.auth_scheme = LDAPAuthProtocol.SICILY
				self.__pwpreprocess = 'HEX'

			if x == PLAINTEXTSCHEME.SICILY_B64:
				self.auth_scheme = LDAPAuthProtocol.SICILY
				self.__pwpreprocess = 'B64'

			if x == PLAINTEXTSCHEME.NTLM_PROMPT:
				self.auth_scheme = LDAPAuthProtocol.NTLM_PASSWORD
				self.__pwpreprocess = 'PROMPT'

			if x == PLAINTEXTSCHEME.NTLM_HEX:
				self.auth_scheme = LDAPAuthProtocol.NTLM_PASSWORD
				self.__pwpreprocess = 'HEX'

			if x == PLAINTEXTSCHEME.NTLM_B64:
				self.auth_scheme = LDAPAuthProtocol.NTLM_PASSWORD
				self.__pwpreprocess = 'B64'			
		except:
			try:
				self.auth_scheme = LDAPAuthProtocol(schemes[1])
			except:
				raise Exception('Uknown scheme!')
		
		return

	def parse(self):
		url_e = urlparse(self.url)
		self.scheme_decoder(url_e.scheme)

		self.password = url_e.password
		if self.__pwpreprocess is not None:
			if self.__pwpreprocess == 'PROMPT':
				self.password = getpass.getpass()

			elif self.__pwpreprocess == 'HEX':
				self.password = bytes.fromhex(self.password).decode()

			elif self.__pwpreprocess == 'B64':
				self.password = base64.b64decode(self.password).decode()

			else:
				raise Exception('Unknown password preprocess directive %s' % self.__pwpreprocess)


		if url_e.username is not None:
			if url_e.username.find('\\') != -1:
				self.domain , self.username = url_e.username.split('\\')
			else:
				self.domain = None
				self.username = url_e.username

		#defaulting schemes...
		if self.auth_scheme is None:
			if self.username is not None and self.domain is not None and self.password is not None:
				#tricky parsing to make user feel confortable...
				if len(self.password) == 32:
					try:
						bytes.fromhex(self.password)
						self.auth_scheme = LDAPAuthProtocol.NTLM_NT
					except:
						self.auth_scheme = LDAPAuthProtocol.NTLM_PASSWORD
				else:
					self.auth_scheme = LDAPAuthProtocol.NTLM_PASSWORD
			else:
				self.auth_scheme = LDAPAuthProtocol.SIMPLE

		self.ldap_host = url_e.hostname
		if url_e.port is not None:
			self.ldap_port = int(url_e.port)

		if url_e.path is not None:
			tree = url_e.path.replace('/','')
			if tree != '':
				self.ldap_tree = tree

		proxy_present = False
		if url_e.query is not None:
			query = parse_qs(url_e.query)
			if 'etype' in query:
				self.etypes = []
			for k in query:
				if k.startswith('proxy') is True:
					proxy_present = True
				if k == 'dc':
					self.dc_ip = query[k][0]
				elif k == 'timeout':
					self.timeout = int(query[k][0])
				elif k == 'serverip':
					self.serverip = query[k][0]
				elif k == 'dns':
					self.dns = query[k] #multiple dns can be set, so not trimming here
				elif k == 'encrypt':
					self.encrypt = bool(int(query[k][0]))
				elif k == 'etype':
					self.etypes = [int(x) for x in query[k]]
				elif k.startswith('auth'):
					self.auth_settings[k[len('auth'):]] = query[k]
				elif k == 'rate':
					self.target_ratelimit = float(query[k][0])
				elif k == 'pagesize':
					self.target_pagesize = int(query[k][0])
				elif k == 'altname':
					self.altname = query[k][0]
				elif k == 'altdomain':
					self.altdomain = query[k][0]
				#elif k.startswith('same'):
				#	self.auth_settings[k[len('same'):]] = query[k]

		if proxy_present is True:
			self.proxy = MSLDAPProxy.from_params(self.url)

		if self.auth_scheme in [LDAPAuthProtocol.SSPI_NTLM, LDAPAuthProtocol.SSPI_KERBEROS]:
			if platform.system().upper() != 'WINDOWS':
				raise Exception('SSPI auth only works on Windows!')
			if self.username is None:
				self.username = '<CURRENT>'
			if self.password is None:
				self.password = '<CURRENT>'
			if self.domain is None:
				self.domain = '<CURRENT>'

		if self.auth_scheme in MSLDAP_KERBEROS_PROTOCOLS and self.dc_ip is None:
			raise Exception('The "dc" parameter MUST be used for kerberos authentication types!')


#		if self.proxy_scheme in [LDAPProxyType.MULTIPLEXOR, LDAPProxyType.MULTIPLEXOR_SSL]:
#			if self.proxy_port is None:
#				self.proxy_port = 9999
#		
#		if self.proxy_scheme in [LDAPProxyType.MULTIPLEXOR, LDAPProxyType.MULTIPLEXOR_SSL]:
#			if 'agentid' not in self.proxy_settings:
#				raise Exception('multiplexor proxy reuires agentid to be set! Set it via proxyagentid parameter!')
#



		
if __name__ == '__main__':
	url_tests = [
		'ldap://10.10.10.2',
		'ldap://10.10.10.2:9999',
		'ldap://test:password@10.10.10.2',
		'ldap://domain\\test@10.10.10.2',
		'ldap://domain\\test:password@10.10.10.2:9999',
		'ldap://domain\\test:AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA@10.10.10.2:9999',
		'ldaps+sspi-ntlm://10.10.10.2',
		'ldaps+sspi-kerberos://10.10.10.2',
		'ldaps+ntlm-password://domain\\test:password@10.10.10.2:9999',
		'ldaps+ntlm-nt://domain\\test:password@10.10.10.2:9999',
		'ldaps+kerberos-password://domain\\test:password@10.10.10.2:9999',
		'ldaps://10.10.10.2:9999',
		'ldaps://test:password@10.10.10.2',
		'ldaps://domain\\test@10.10.10.2',
		'ldaps://domain\\test:password@10.10.10.2:9999',
		'ldaps://DOMAIN\\test:password@10.10.10.2:9999/?proxytype=socks5&proxyserver=127.0.0.1',
		'ldaps://DOMAIN\\test:password@10.10.10.2:9999/?proxytype=socks5&proxyserver=127.0.0.1&proxyuser=admin&proxypass=alma',
		'ldaps://DOMAIN\\test:password@10.10.10.2:9999/?proxytype=multiplexor&proxyserver=127.0.0.1&proxyport=9999&proxyuser=admin&proxypass=alma',
		'ldaps://10.10.10.2',
		'ldaps://10.10.10.2:6666',
	]
	for url in url_tests:
		print('===========================================================================')
		print(url)
		try:
			dec = MSLDAPURLDecoder(url)
			creds = dec.get_credential()
			target = dec.get_target()
		except Exception as e:
			import traceback
			traceback.print_exc()
			print('ERROR! Reason: %s' % e)
			input()
		else:
			print(str(creds))
			print(str(target))
			input()




## 
##
## Interface to allow remote kerberos authentication via Multiplexor
## 
##
##
##
##
## TODO: RPC auth type is not implemented or tested!!!!

from msldap.authentication.spnego.asn1_structs import KRB5Token
from msldap.authentication.kerberos.gssapi import get_gssapi, GSSWrapToken, KRB5_MECH_INDEP_TOKEN
from minikerberos.protocol.asn1_structs import AP_REQ, AP_REP, TGS_REP
from minikerberos.protocol.encryption import Enctype, Key, _enctype_table

from multiplexor.operator.external.sspi import KerberosSSPIClient
from multiplexor.operator import MultiplexorOperator
import enum

# mutual auth not supported
# encryption is always on
# we dont get the output flags back (lack of time to do the multiplexor protocol... TODO

class ISC_REQ(enum.IntFlag):
	DELEGATE = 1
	MUTUAL_AUTH = 2
	REPLAY_DETECT = 4
	SEQUENCE_DETECT = 8
	CONFIDENTIALITY = 16
	USE_SESSION_KEY = 32
	PROMPT_FOR_CREDS = 64
	USE_SUPPLIED_CREDS = 128
	ALLOCATE_MEMORY = 256
	USE_DCE_STYLE = 512
	DATAGRAM = 1024
	CONNECTION = 2048
	CALL_LEVEL = 4096
	FRAGMENT_SUPPLIED = 8192
	EXTENDED_ERROR = 16384
	STREAM = 32768
	INTEGRITY = 65536
	IDENTIFY = 131072
	NULL_SESSION = 262144
	MANUAL_CRED_VALIDATION = 524288
	RESERVED1 = 1048576
	FRAGMENT_TO_FIT = 2097152
	HTTP = 0x10000000

class MSLDAPKerberosMultiplexor:
	def __init__(self, settings):
		self.iterations = 0
		self.settings = settings
		self.mode = 'CLIENT'
		self.ksspi = None
		self.client = None
		self.target = None
		self.gssapi = None
		self.etype = None
		self.session_key = None
		self.seq_number = 0
		self.flags = ISC_REQ.CONNECTION
		
		self.setup()
		
	def setup(self):
		if self.settings.encrypt is True:
			self.flags = \
				ISC_REQ.CONFIDENTIALITY |\
				ISC_REQ.INTEGRITY |\
				ISC_REQ.REPLAY_DETECT |\
				ISC_REQ.SEQUENCE_DETECT

	def get_seq_number(self):
		"""
		Fetches the starting sequence number. This is either zero or can be found in the authenticator field of the 
		AP_REQ structure. As windows uses a random seq number AND a subkey as well, we can't obtain it by decrypting the 
		AP_REQ structure. Insead under the hood we perform an encryption operation via EncryptMessage API which will 
		yield the start sequence number
		"""
		return self.seq_number
		
	async def encrypt(self, data, message_no):
		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction, auth_data=auth_data)
	
	def signing_needed(self):
		"""
		Checks if integrity protection was negotiated
		"""
		return ISC_REQ.INTEGRITY in self.flags

	def encryption_needed(self):
		"""
		Checks if confidentiality flag was negotiated
		"""
		return ISC_REQ.CONFIDENTIALITY in self.flags

	def get_session_key(self):
		return self.session_key
	
	async def authenticate(self, authData = None, flags = None, seq_number = 0, cb_data=None):
		#authdata is only for api compatibility reasons
		if self.ksspi is None:
			await self.start_remote_kerberos()
		try:
				apreq, res = await self.ksspi.authenticate(self.settings.target.to_target_string(), flags=str(self.flags.value))
				#print('MULTIPLEXOR KERBEROS SSPI, APREQ: %s ERROR: %s' % (apreq, res))
				if res is not None:
					return None, None, res

				# here it seems like we get the full token not just the apreq data...
				# so we need to discard the layers

				self.session_key, err = await self.ksspi.get_session_key()
				if err is not None:
					return None, None, err

				unwrap = KRB5_MECH_INDEP_TOKEN.from_bytes(apreq)
				aprep = AP_REQ.load(unwrap.data[2:]).native
				subkey = Key(aprep['ticket']['enc-part']['etype'], self.session_key)
				self.gssapi = get_gssapi(subkey)
				
				if aprep['ticket']['enc-part']['etype'] != 23:
					raw_seq_data, err = await self.ksspi.get_seq_number()
					if err is not None:
						return None, None, err
					self.seq_number = GSSWrapToken.from_bytes(raw_seq_data[16:]).SND_SEQ
				
				return unwrap.data[2:], False, res
		except Exception as e:
			return None, None, e

	async def start_remote_kerberos(self):
		try:
			#print(self.settings.get_url())
			#print(self.settings.agent_id)
			self.operator = MultiplexorOperator(self.settings.get_url())
			await self.operator.connect()
			#creating virtual sspi server
			server_info = await self.operator.start_sspi(self.settings.agent_id)
			#print(server_info)

			sspi_url = 'ws://%s:%s' % (server_info['listen_ip'], server_info['listen_port'])

			#print(sspi_url)
			self.ksspi = KerberosSSPIClient(sspi_url)
			await self.ksspi.connect()
		except Exception as e:
			import traceback
			traceback.print_exc()
			return None
		



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import enum
from urllib.parse import urlparse, parse_qs

from asysocks.common.clienturl import SocksClientURL 

class MSLDAPProxyType(enum.Enum):
	SOCKS4 = 'SOCKS4'
	SOCKS4_SSL = 'SOCKS4_SSL'
	SOCKS5 = 'SOCKS5'
	SOCKS5_SSL = 'SOCKS5_SSL'
	MULTIPLEXOR = 'MULTIPLEXOR'
	MULTIPLEXOR_SSL = 'MULTIPLEXOR_SSL'
	WSNET = 'WSNET'
	WSNETWS = 'WSNETWS'
	WSNETWSS = 'WSNETWSS'

class MSLDAPProxy:
	"""
	Describes the proxy to be used when connecting to the server. Used as a parameter to the `MSLDAPTarget` object
	
	:param type: Specifies the proxy type
	:type type: :class:`MSLDAPProxyType`
	:param target: 
	:type target: 
	:param auth: Specifies the proxy authentication if any
	:type auth: 
	"""
	def __init__(self, type = None, target = None, auth = None):
		self.type = type
		self.target = target
		self.auth = auth


	@staticmethod
	def from_params(url_str):
		"""
		Creates a proxy object from the parameters found in an LDAP URL string

		:param type: url_str
		:type type: str
		:return: The proxy object 
		:rtype: :class:`MSLDAPProxy`
		"""
		proxy = MSLDAPProxy()
		url = urlparse(url_str)
		if url.query is None:
			return None

		query = parse_qs(url.query)
		if 'proxytype' not in query and 'sametype' not in query:
			return None
		
		proxy.type = MSLDAPProxyType(query['proxytype'][0].upper())
		if proxy.type in [MSLDAPProxyType.WSNET, MSLDAPProxyType.WSNETWS, MSLDAPProxyType.WSNETWSS,MSLDAPProxyType.SOCKS4, MSLDAPProxyType.SOCKS4_SSL, MSLDAPProxyType.SOCKS5, MSLDAPProxyType.SOCKS5_SSL]:
			proxy.target = SocksClientURL.from_params(url_str)
		else:
			proxy.target  = MSLDAPMultiplexorProxy.from_params(url_str)
		
		return proxy

	def __str__(self):
		t = '==== MSLDAPProxy ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t


class MSLDAPMultiplexorProxy:
	def __init__(self):
		self.ip = None
		self.port = None
		self.timeout = 10
		self.type = MSLDAPProxyType.MULTIPLEXOR
		self.username = None
		self.password = None
		self.domain = None
		self.agent_id = None
		self.virtual_socks_port = None
		self.virtual_socks_ip = None
	
	def sanity_check(self):
		if self.ip is None:
			raise Exception('MULTIPLEXOR server IP is missing!')
		if self.port is None:
			raise Exception('MULTIPLEXOR server port is missing!')
		if self.agent_id is None:
				raise Exception('MULTIPLEXOR proxy requires agentid to be set!')

	def get_server_url(self):
		con_str = 'ws://%s:%s' % (self.ip, self.port)
		if self.type == MSLDAPProxyType.MULTIPLEXOR_SSL:
			con_str = 'wss://%s:%s' % (self.ip, self.port)
		return con_str

	@staticmethod
	def from_params(url_str):
		res = MSLDAPMultiplexorProxy()
		url = urlparse(url_str)
		res.endpoint_ip = url.hostname
		if url.port:
			res.endpoint_port = int(url.port)
		if url.query is not None:
			query = parse_qs(url.query)

			for k in query:
				if k.startswith('proxy'):
					if k[5:] in multiplexorproxyurl_param2var:

						data = query[k][0]
						for c in multiplexorproxyurl_param2var[k[5:]][1]:
							data = c(data)

						setattr(
							res, 
							multiplexorproxyurl_param2var[k[5:]][0], 
							data
						)
		res.sanity_check()

		return res

def stru(x):
	return str(x).upper()

multiplexorproxyurl_param2var = {
	'type' : ('version', [stru, MSLDAPProxyType]),
	'host' : ('ip', [str]),
	'port' : ('port', [int]),
	'timeout': ('timeout', [int]),
	'user' : ('username', [str]),
	'pass' : ('password', [str]),
	#'authtype' : ('authtype', [SOCKS5Method]),
	'agentid' : ('agent_id', [str]),
	'domain' : ('domain', [str])

}




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import asyncio
import traceback
import logging
import csv
import shlex
import datetime

from msldap.external.aiocmd.aiocmd import aiocmd
from msldap.external.asciitree.asciitree import LeftAligned
from tqdm import tqdm

from msldap import logger
from asysocks import logger as sockslogger
from msldap.client import MSLDAPClient
from msldap.commons.url import MSLDAPURLDecoder
from msldap.ldap_objects import MSADUser, MSADMachine, MSADUser_TSV_ATTRS

from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR


class MSLDAPCompDomainList:
	def __init__(self, ldap_url):
		self.conn_url = ldap_url
		self.connection = None
		self.adinfo = None
		self.ldapinfo = None
		self.domain_name = None

	async def login(self):
		"""Performs connection and login"""
		try:
			logger.debug(self.conn_url.get_credential())
			logger.debug(self.conn_url.get_target())
			
			
			self.connection = self.conn_url.get_client()
			_, err = await self.connection.connect()
			if err is not None:
				raise err
			
			return True, None
		except Exception as e:
			return False, e

	async def do_adinfo(self, show = True):
		"""Prints detailed Active Driectory info"""
		try:
			if self.adinfo is None:
				self.adinfo = self.connection._ldapinfo
				self.domain_name = self.adinfo.distinguishedName.replace('DC','').replace('=','').replace(',','.')
			if show is True:
				print(self.adinfo)
			
			return True, None
		except Exception as e:
			return False, e

	async def run(self):
		try:
			_, err = await self.login()
			if err is not None:
				raise err
			_, err = await self.do_adinfo(False)
			if err is not None:
				raise err
			#machine_filename = '%s_computers_%s.txt' % (self.domain_name, datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
		
			async for machine, err in self.connection.get_all_machines(attrs=['sAMAccountName', 'dNSHostName']):
				if err is not None:
					raise err
					
				dns = machine.dNSHostName
				if dns is None:
					dns = '%s.%s' % (machine.sAMAccountName[:-1], self.domain_name)

				print(str(dns))
		except:
			traceback.print_exc()


def main():
	import argparse
	parser = argparse.ArgumentParser(description='MS LDAP library')
	parser.add_argument('-v', '--verbose', action='count', default=0, help='Verbosity, can be stacked')
	parser.add_argument('-n', '--no-interactive', action='store_true')
	parser.add_argument('url', help='Connection string in URL format.')

	args = parser.parse_args()
	
	
	###### VERBOSITY
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	else:
		sockslogger.setLevel(logging.DEBUG)
		logger.setLevel(logging.DEBUG)
		logging.basicConfig(level=logging.DEBUG)

	ldap_url = MSLDAPURLDecoder(args.url)
	compdomlist = MSLDAPCompDomainList(ldap_url)


	asyncio.run(compdomlist.run())

if __name__ == '__main__':
	main()




from msldap.commons.utils import print_cert
from asn1crypto.x509 import Certificate

MSADEnrollmentService_ATTRS = ['cACertificate', 'msPKI-Enrollment-Servers', 'dNSHostName', 'cn', 'sn', 'distinguishedName', 'whenChanged', 'whenCreated', 'name', 'displayName', 'cACertificateDN', 'certificateTemplates']

class MSADEnrollmentService:
	def __init__(self):
		self.sn = None #str
		self.cn = None #str
		self.distinguishedName = None #dn
		self.name = None
		self.displayName = None
		self.cACertificate = None
		self.cACertificateDN = None
		self.dNSHostName = None
		self.certificateTemplates = []
		self.enrollmentServers = []
		
	@staticmethod
	def from_ldap(entry):
		adi = MSADEnrollmentService()
		adi.sn = entry['attributes'].get('sn') 
		adi.cn = entry['attributes'].get('cn') 
		adi.distinguishedName = entry['attributes'].get('distinguishedName')
		adi.cACertificate = entry['attributes'].get('cACertificate')
		if adi.cACertificate is not None:
			adi.cACertificate = Certificate.load(adi.cACertificate)
		adi.name = entry['attributes'].get('name')
		adi.displayName = entry['attributes'].get('displayName')
		adi.dNSHostName = entry['attributes'].get('dNSHostName')
		adi.cACertificateDN = entry['attributes'].get('cACertificateDN')
		adi.certificateTemplates = entry['attributes'].get('certificateTemplates')
		for serverdef in entry['attributes'].get('msPKI-Enrollment-Servers', []):
			adi.enrollmentServers.append(serverdef.split('\n')[3])
		return adi
		
	def __str__(self):
		t = '== MSADEnrollmentService ==\r\n'
		t += "Name: %s\r\n" % self.name
		t += "DNS name: %s\r\n" % self.dNSHostName
		t += "Templates: %s\r\n" % ', '.join(self.certificateTemplates)
		if len(self.enrollmentServers) > 0:
			t += "Web services: %s\r\n" % ", ".join(self.enrollmentServers)
		t += "Certificate: \r\n%s\r\n" % print_cert(self.cACertificate.native, 2)

		return t


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import enum

# https://tools.ietf.org/html/rfc4513
# simple auth: 
# 	- anonymous
#   - user without password
# 	- username + password
#
# SASL:
#   - plain
#   - gssapi
#		-SSPI
#			- NTLM
#			- KERBEROS
# Sicily:
#   - NTLM
# Multiplexor
#		

class LDAPAuthProtocol(enum.Enum):
	SIMPLE = 'SIMPLE' #SIMPLE can be with no creds - anonymous bind
	PLAIN = 'PLAIN' #actually SASL-PLAIN
	SICILY = 'SICILY' #NTLM (old proprietary from MS)
	NTLM_PASSWORD = 'NTLM_PASSWORD' #actually SASL-GSSAPI-SPNEGO-NTLM
	NTLM_NT = 'NTLM_NT' #actually SASL-GSSAPI-SPNEGO-NTLM
	KERBEROS_RC4 = 'KERBEROS_RC4' 
	KERBEROS_NT = 'KERBEROS_NT' 
	KERBEROS_AES = 'KERBEROS_AES' 
	KERBEROS_PASSWORD = 'KERBEROS_PASSWORD' 
	KERBEROS_CCACHE = 'KERBEROS_CCACHE' 
	KERBEROS_KEYTAB = 'KERBEROS_KEYTAB'
	KERBEROS_KIRBI = 'KERBEROS_KIRBI'
	KERBEROS_PFX = 'KERBEROS_PFX'
	KERBEROS_PEM = 'KERBEROS_PEM'
	KERBEROS_CERTSTORE = 'KERBEROS_CERTSTORE'
	MULTIPLEXOR_KERBEROS = 'MULTIPLEXOR_KERBEROS'
	MULTIPLEXOR_NTLM = 'MULTIPLEXOR_NTLM'
	MULTIPLEXOR_SSL_KERBEROS = 'MULTIPLEXOR_SSL_KERBEROS'
	MULTIPLEXOR_SSL_NTLM = 'MULTIPLEXOR_SSL_NTLM'
	SSPI_NTLM = 'SSPI_NTLM' #actually SASL-GSSAPI-SPNEGO-NTLM but with integrated SSPI
	SSPI_KERBEROS = 'SSPI_KERBEROS' #actually SASL-GSSAPI-SPNEGO-KERBEROS but with integrated SSPI
	WSNET_NTLM = 'WSNET_NTLM'
	WSNET_KERBEROS = 'WSNET_KERBEROS'
	SSPIPROXY_NTLM = 'SSPIPROXY_NTLM'
	SSPIPROXY_KERBEROS = 'SSPIPROXY_KERBEROS'

MSLDAP_GSS_METHODS = [
		LDAPAuthProtocol.NTLM_PASSWORD ,
		LDAPAuthProtocol.NTLM_NT ,
		LDAPAuthProtocol.KERBEROS_RC4 ,
		LDAPAuthProtocol.KERBEROS_NT ,
		LDAPAuthProtocol.KERBEROS_AES ,
		LDAPAuthProtocol.KERBEROS_PASSWORD ,
		LDAPAuthProtocol.KERBEROS_CCACHE ,
		LDAPAuthProtocol.KERBEROS_KEYTAB ,
		LDAPAuthProtocol.KERBEROS_KIRBI ,
		LDAPAuthProtocol.KERBEROS_PFX ,
		LDAPAuthProtocol.KERBEROS_PEM ,
		LDAPAuthProtocol.KERBEROS_CERTSTORE ,
		LDAPAuthProtocol.SSPI_NTLM ,
		LDAPAuthProtocol.SSPI_KERBEROS,
		LDAPAuthProtocol.MULTIPLEXOR_KERBEROS,
		LDAPAuthProtocol.MULTIPLEXOR_NTLM,
		LDAPAuthProtocol.MULTIPLEXOR_SSL_KERBEROS,
		LDAPAuthProtocol.MULTIPLEXOR_SSL_NTLM,
		LDAPAuthProtocol.WSNET_NTLM,
		LDAPAuthProtocol.WSNET_KERBEROS,
		LDAPAuthProtocol.SSPIPROXY_NTLM,
		LDAPAuthProtocol.SSPIPROXY_KERBEROS,
]

MSLDAP_KERBEROS_PROTOCOLS = [
	LDAPAuthProtocol.KERBEROS_RC4 ,
	LDAPAuthProtocol.KERBEROS_NT ,
	LDAPAuthProtocol.KERBEROS_AES ,
	LDAPAuthProtocol.KERBEROS_PASSWORD ,
	LDAPAuthProtocol.KERBEROS_CCACHE ,
	LDAPAuthProtocol.KERBEROS_KEYTAB ,
	LDAPAuthProtocol.KERBEROS_KIRBI ,
	LDAPAuthProtocol.KERBEROS_PFX ,
	LDAPAuthProtocol.KERBEROS_PEM ,
	LDAPAuthProtocol.KERBEROS_CERTSTORE ,
]

class MSLDAPCredential:
	"""
	Describes the user's credentials to be used for authentication during the bind operation.
	
	:param domain: Domain of the user
	:type domain: str
	:param username: Username of the user
	:type username: str
	:param password: The authentication secret. The actual contents depend on the `auth_method`
	:type password: str
	:param auth_method: The ahtentication method to be performed during bind operation
	:type auth_method: :class:`LDAPAuthProtocol`
	:param settings: Additional settings
	:type settings: dict
	:param etypes: Supported encryption types for Kerberos authentication.
	:type etypes: List[:class:`int`]
	:param encrypt: Use protocol-level encryption. Doesnt work on LDAPS
	:type encrypt: bool
	"""
	def __init__(self, domain=None, username= None, password = None, auth_method = None, settings = None, etypes = None, encrypt = False, altname = None, altdomain = None):
		self.auth_method = auth_method
		self.domain   = domain
		self.username = username
		self.password = password
		self.signing_preferred = False
		self.encryption_preferred = False
		self.settings = settings
		self.etypes = etypes
		self.encrypt = encrypt
		self.altname = altname
		self.altdomain = altdomain

	def get_msuser(self):
		if not self.domain:
			return self.username

		return '%s\\%s' % (self.domain,self.username)		

	def __str__(self):
		t = '==== MSLDAPCredential ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t




import enum
import platform

import copy
from msldap.commons.credential import MSLDAPCredential, LDAPAuthProtocol
from msldap.authentication.spnego.native import SPNEGO
from msldap.authentication.ntlm.native import NTLMAUTHHandler, NTLMHandlerSettings
from msldap.authentication.kerberos.native import MSLDAPKerberos
from msldap.commons.proxy import MSLDAPProxyType
from minikerberos.common.target import KerberosTarget
from minikerberos.common.proxy import KerberosProxy
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.spn import KerberosSPN

from minikerberos.network.selector import KerberosClientSocketSelector


if platform.system().upper() == 'WINDOWS':
	from msldap.authentication.kerberos.sspi import MSLDAPKerberosSSPI
	from msldap.authentication.ntlm.sspi import MSLDAPNTLMSSPI

class MSLDAPNTLMCredential:
	def __init__(self):
		self.username = None
		self.domain = ''
		self.password = None
		self.workstation = None
		self.is_guest = False
		self.nt_hash = None
		self.lm_hash = None 
		self.encrypt = False

class MSLDAPSIMPLECredential:
	def __init__(self):
		self.username = None
		self.domain = None
		self.password = None

class MSLDAPPLAINCredential:
	def __init__(self):
		self.username = None
		self.domain = None
		self.password = None

class MSLDAPKerberosCredential:
	def __init__(self):
		self.connection = None #KerberosCredential
		self.target = None #KerberosTarget
		self.ksoc = None #KerberosSocketAIO
		self.ccred = None
		self.encrypt = False
		self.enctypes = None #[23,17,18]

class MSLDAPKerberosSSPICredential:
	def __init__(self):
		self.domain = None
		self.password = None
		self.username  = None
		self.encrypt = False
		
class MSLDAPNTLMSSPICredential:
	def __init__(self):
		self.username = None
		self.password = None
		self.domain = None
		self.encrypt = False

class MSLDAPWSNETCredential:
	def __init__(self):
		self.type = 'NTLM'
		self.username = '<CURRENT>'
		self.domain = '<CURRENT>'
		self.password = '<CURRENT>'
		self.target = None
		self.is_guest = False
		self.agent_id = None
		self.encrypt = False

class MSLDAPSSPIProxyCredential:
	def __init__(self):
		self.type = 'NTLM'
		self.username = '<CURRENT>'
		self.domain = '<CURRENT>'
		self.password = '<CURRENT>'
		self.target = None
		self.is_guest = False
		self.agent_id = None
		self.encrypt = False
		self.host = '127.0.0.1'
		self.port = 9999
		self.proto = 'ws'

		

class MSLDAPMultiplexorCredential:
	def __init__(self):
		self.type = 'NTLM'
		self.username = '<CURRENT>'
		self.domain = '<CURRENT>'
		self.password = '<CURRENT>'
		self.target = None
		self.is_guest = False
		self.is_ssl = False
		self.mp_host = '127.0.0.1'
		self.mp_port = 9999
		self.mp_username = None
		self.mp_domain = None
		self.mp_password = None
		self.agent_id = None
		self.encrypt = False

	def get_url(self):
		url_temp = 'ws://%s:%s'
		if self.is_ssl is True:
			url_temp = 'wss://%s:%s'
		url = url_temp % (self.mp_host, self.mp_port)
		return url

	def parse_settings(self, settings):
		req = ['agentid']
		for r in req:
			if r not in settings:
				raise Exception('%s parameter missing' % r)
		self.mp_host = settings.get('host', ['127.0.0.1'])[0]
		self.mp_port = settings.get('port', ['9999'])[0]
		if self.mp_port is None:
			self.mp_port = '9999'
		if 'user' in settings:
			self.mp_username = settings.get('user')[0]
		if 'domain' in settings:
			self.mp_domain = settings.get('domain')[0]
		if 'password' in settings:
			self.mp_password = settings.get('password')[0]
		self.agent_id = settings['agentid'][0]



"""
class LDAPAuthProtocol(enum.Enum):
	PLAIN = 'PLAIN' #actually SASL-PLAIN

	MULTIPLEXOR = 'MULTIPLEXOR'
	MULTIPLEXOR_SSL = 'MULTIPLEXOR_SSL'
	SSPI_NTLM = 'SSPI_NTLM' #actually SASL-GSSAPI-SPNEGO-NTLM but with integrated SSPI
	SSPI_KERBEROS = 'SSPI_KERBEROS' #actually SASL-GSSAPI-SPNEGO-KERBEROS but with integrated SSPI
"""

class AuthenticatorBuilder:
	def __init__(self, creds, target = None):
		self.creds = creds
		self.target = target
	
	def build(self):
		if self.creds.auth_method == LDAPAuthProtocol.SICILY:
			ntlmcred = MSLDAPNTLMCredential()
			ntlmcred.username = self.creds.username
			ntlmcred.domain = self.creds.domain if self.creds.domain is not None else ''
			ntlmcred.workstation = None
			ntlmcred.is_guest = False
			ntlmcred.encrypt = self.creds.encrypt

			
			if self.creds.password is None:
				raise Exception('NTLM authentication requres password/NT hash!')
			
			
			if len(self.creds.password) == 32:
				try:
					bytes.fromhex(self.creds.password)
				except:
					ntlmcred.password = self.creds.password
				else:
					ntlmcred.nt_hash = self.creds.password
			
			else:
				ntlmcred.password = self.creds.password
			
			settings = NTLMHandlerSettings(ntlmcred)
			return NTLMAUTHHandler(settings)

		elif self.creds.auth_method == LDAPAuthProtocol.SIMPLE:
			cred = MSLDAPPLAINCredential()
			cred.username = self.creds.username
			cred.domain = self.creds.domain
			cred.password = self.creds.password
			return cred

		elif self.creds.auth_method == LDAPAuthProtocol.PLAIN:
			cred = MSLDAPSIMPLECredential()
			cred.username = self.creds.username
			cred.domain = self.creds.domain
			cred.password = self.creds.password
			return cred

		elif self.creds.auth_method in [LDAPAuthProtocol.NTLM_PASSWORD, LDAPAuthProtocol.NTLM_NT]:
			ntlmcred = MSLDAPNTLMCredential()
			ntlmcred.username = self.creds.username
			ntlmcred.domain = self.creds.domain if self.creds.domain is not None else ''
			ntlmcred.workstation = None
			ntlmcred.is_guest = False
			ntlmcred.encrypt = self.creds.encrypt
			
			if self.creds.password is None:
				raise Exception('NTLM authentication requres password!')

			if self.creds.auth_method == LDAPAuthProtocol.NTLM_PASSWORD:
				ntlmcred.password = self.creds.password
			elif self.creds.auth_method == LDAPAuthProtocol.NTLM_NT:
				ntlmcred.nt_hash = self.creds.password
			else:
				raise Exception('Unknown NTLM auth method!')
			
			settings = NTLMHandlerSettings(ntlmcred)
			handler = NTLMAUTHHandler(settings)
			
			##setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
			
			return spneg

		elif self.creds.auth_method in [
				LDAPAuthProtocol.KERBEROS_RC4, 
				LDAPAuthProtocol.KERBEROS_NT, 
				LDAPAuthProtocol.KERBEROS_AES,
				LDAPAuthProtocol.KERBEROS_PASSWORD, 
				LDAPAuthProtocol.KERBEROS_CCACHE, 
				LDAPAuthProtocol.KERBEROS_KEYTAB,
				LDAPAuthProtocol.KERBEROS_KIRBI,
				LDAPAuthProtocol.KERBEROS_PFX,
				LDAPAuthProtocol.KERBEROS_PEM,
				LDAPAuthProtocol.KERBEROS_CERTSTORE]:
			
			if self.target is None:
				raise Exception('Target must be specified with Kerberos!')
				
			if self.target.host is None:
				raise Exception('target must have a domain name or hostname for kerberos!')
				
			if self.target.dc_ip is None:
				raise Exception('target must have a dc_ip for kerberos!')
			
			kcred = MSLDAPKerberosCredential()
			if self.creds.auth_method == LDAPAuthProtocol.KERBEROS_KIRBI:
				kc = KerberosCredential.from_kirbi(self.creds.password, self.creds.username, self.creds.domain)
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_CCACHE:
				kc = KerberosCredential.from_ccache_file(self.creds.password, self.creds.username, self.creds.domain)
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_KEYTAB:
				kc = KerberosCredential.from_kirbi(self.creds.password, self.creds.username, self.creds.domain)
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_PFX:
				kc = KerberosCredential.from_pfx_file(self.creds.username, self.creds.password, username = self.creds.altname, domain = self.creds.altdomain)
				self.creds.username = kc.username
				self.creds.domain = kc.domain
				self.target.domain = kc.domain
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_PEM:
				kc = KerberosCredential.from_pem_file(self.creds.username, self.creds.password, username = self.creds.altname, domain = self.creds.altdomain)
				self.creds.username = kc.username
				self.creds.domain = kc.domain
				self.target.domain = kc.domain
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_CERTSTORE:
				# username is the CN of the certificate
				# secret is the name of the certstore, default: MY
				certstore = self.creds.secret
				if self.creds.secret is None:
					certstore = 'MY'
				kc = KerberosCredential.from_windows_certstore(self.creds.username, certstore, username = self.creds.altname, domain = self.creds.altdomain)
				self.creds.username = kc.username
				self.creds.domain = kc.domain
				self.target.domain = kc.domain

			else:
				kc = KerberosCredential()
				kc.username = self.creds.username
				kc.domain = self.creds.domain
			kcred.enctypes = []
			if self.creds.auth_method == LDAPAuthProtocol.KERBEROS_PASSWORD:
				kc.password = self.creds.password
				kcred.enctypes = [23,17,18]
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_NT:
				kc.nt_hash = self.creds.password
				kcred.enctypes = [23]
				
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_AES:
				if len(self.creds.password) == 32:
					kc.kerberos_key_aes_128 = self.creds.password
					kcred.enctypes = [17]
				elif len(self.creds.password) == 64:
					kc.kerberos_key_aes_256 = self.creds.password
					kcred.enctypes = [18]
					
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_RC4:
				kc.kerberos_key_rc4 = self.creds.password
				kcred.enctypes = [23]
			
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_CCACHE:
				kcred.enctypes = [23,17,18] # TODO: fix this
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_KEYTAB:
				kc.keytab = self.creds.password
				kcred.enctypes = [23,17,18] # TODO: fix this
			elif self.creds.auth_method == LDAPAuthProtocol.KERBEROS_KIRBI:
				kcred.enctypes = [23,17,18] # TODO: fix this
			elif self.creds.auth_method in [LDAPAuthProtocol.KERBEROS_PFX, LDAPAuthProtocol.KERBEROS_CERTSTORE, LDAPAuthProtocol.KERBEROS_PEM]:
				kcred.enctypes = [17,18]
			else:
				raise Exception('No suitable secret type found to set up kerberos!')

			if self.creds.etypes is not None:
				kcred.enctypes = list(set(self.creds.etypes).intersection(set(kcred.enctypes)))				
			
			kcred.ccred = kc
			kcred.spn = KerberosSPN.from_target_string(self.target.to_target_string())
			kcred.target = KerberosTarget(self.target.dc_ip)
			kcred.encrypt = self.creds.encrypt
			
			if self.target.proxy is not None:
					kcred.target.proxy = KerberosProxy()
					kcred.target.proxy.type = self.target.proxy.type
					kcred.target.proxy.target = copy.deepcopy(self.target.proxy.target)
					kcred.target.proxy.target[-1].endpoint_ip = self.target.dc_ip
					kcred.target.proxy.target[-1].endpoint_port = 88

			handler = MSLDAPKerberos(kcred)
			
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
			return spneg

		elif self.creds.auth_method == LDAPAuthProtocol.SSPI_KERBEROS:
			if self.target is None:
				raise Exception('Target must be specified with Kerberos SSPI!')
				
			kerbcred = MSLDAPKerberosSSPICredential()
			kerbcred.username = self.creds.domain if self.creds.domain is not None else '<CURRENT>'
			kerbcred.username = self.creds.username if self.creds.username is not None else '<CURRENT>'
			kerbcred.password = self.creds.password if self.creds.password is not None else '<CURRENT>'
			kerbcred.spn = self.target.to_target_string()
			kerbcred.encrypt = self.creds.encrypt
			
			handler = MSLDAPKerberosSSPI(kerbcred)
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
			return spneg
		
		elif self.creds.auth_method == LDAPAuthProtocol.SSPI_NTLM:
			ntlmcred = MSLDAPNTLMSSPICredential()
			ntlmcred.username = self.creds.domain if self.creds.domain is not None else '<CURRENT>'
			ntlmcred.username = self.creds.username if self.creds.username is not None else '<CURRENT>'
			ntlmcred.password = self.creds.password if self.creds.password is not None else '<CURRENT>'
			ntlmcred.encrypt = self.creds.encrypt

			handler = MSLDAPNTLMSSPI(ntlmcred)
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
			return spneg

		elif self.creds.auth_method.value.startswith('MULTIPLEXOR'):
			if self.creds.auth_method in [LDAPAuthProtocol.MULTIPLEXOR_SSL_NTLM, LDAPAuthProtocol.MULTIPLEXOR_NTLM]:
				from msldap.authentication.ntlm.multiplexor import MSLDAPNTLMMultiplexor
				ntlmcred = MSLDAPMultiplexorCredential()
				ntlmcred.type = 'NTLM'
				if self.creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if self.creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if self.creds.password is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.is_ssl = True if self.creds.auth_method == LDAPAuthProtocol.MULTIPLEXOR_SSL_NTLM else False
				ntlmcred.parse_settings(self.creds.settings)
				ntlmcred.encrypt = self.creds.encrypt
				
				handler = MSLDAPNTLMMultiplexor(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
				return spneg

			elif self.creds.auth_method in [LDAPAuthProtocol.MULTIPLEXOR_SSL_KERBEROS, LDAPAuthProtocol.MULTIPLEXOR_KERBEROS]:
				from msldap.authentication.kerberos.multiplexor import MSLDAPKerberosMultiplexor

				ntlmcred = MSLDAPMultiplexorCredential()
				ntlmcred.type = 'KERBEROS'
				ntlmcred.target = self.target
				if self.creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if self.creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if self.creds.password is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.is_ssl = True if self.creds.auth_method == LDAPAuthProtocol.MULTIPLEXOR_SSL_NTLM else False
				ntlmcred.parse_settings(self.creds.settings)
				ntlmcred.encrypt = self.creds.encrypt

				handler = MSLDAPKerberosMultiplexor(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
				return spneg

		elif self.creds.auth_method.value.startswith('SSPIPROXY'):
			if self.creds.auth_method == LDAPAuthProtocol.SSPIPROXY_NTLM:
				from msldap.authentication.ntlm.sspiproxy import MSLDAPSSPIProxyNTLMAuth
				ntlmcred = MSLDAPSSPIProxyCredential()
				ntlmcred.type = 'NTLM'
				if self.creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if self.creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if self.creds.password is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.encrypt = self.creds.encrypt
				ntlmcred.host = self.creds.settings['host'][0]
				ntlmcred.port = int(self.creds.settings['port'][0])
				ntlmcred.proto = 'ws'
				if 'proto' in self.creds.settings:
					ntlmcred.proto = self.creds.settings['proto'][0]
				if 'agentid' in self.creds.settings:
					ntlmcred.agent_id = bytes.fromhex(self.creds.settings['agentid'][0])
				
				handler = MSLDAPSSPIProxyNTLMAuth(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
				return spneg

			elif self.creds.auth_method == LDAPAuthProtocol.SSPIPROXY_KERBEROS:
				from msldap.authentication.kerberos.sspiproxyws import MSLDAPSSPIProxyKerberosAuth

				ntlmcred = MSLDAPSSPIProxyCredential()
				ntlmcred.type = 'KERBEROS'
				ntlmcred.target = self.target
				if self.creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if self.creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if self.creds.password is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.encrypt = self.creds.encrypt
				ntlmcred.host = self.creds.settings['host'][0]
				ntlmcred.port = self.creds.settings['port'][0]
				ntlmcred.proto = 'ws'
				if 'proto' in self.creds.settings:
					ntlmcred.proto = self.creds.settings['proto'][0]
				if 'agentid' in self.creds.settings:
					ntlmcred.agent_id = bytes.fromhex(self.creds.settings['agentid'][0])

				handler = MSLDAPSSPIProxyKerberosAuth(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
				return spneg

		elif self.creds.auth_method.value.startswith('WSNET'):
			if self.creds.auth_method in [LDAPAuthProtocol.WSNET_NTLM]:
				from msldap.authentication.ntlm.wsnet import MSLDAPWSNetNTLMAuth
				
				ntlmcred = MSLDAPWSNETCredential()
				ntlmcred.type = 'NTLM'
				if self.creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if self.creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if self.creds.password is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				
				handler = MSLDAPWSNetNTLMAuth(ntlmcred)
				spneg = SPNEGO()
				spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
				return spneg
			

			elif self.creds.auth_method in [LDAPAuthProtocol.WSNET_KERBEROS]:
				from msldap.authentication.kerberos.wsnet import MSLDAPWSNetKerberosAuth

				ntlmcred = MSLDAPWSNETCredential()
				ntlmcred.type = 'KERBEROS'
				ntlmcred.target = self.target
				if self.creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if self.creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if self.creds.password is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False

				handler = MSLDAPWSNetKerberosAuth(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
				return spneg


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import enum

from winacl.dtyp.sid import SID
from msldap.ldap_objects.common import MSLDAP_UAC, vn

MSADDomainTrust_ATTRS = [ 
	'sn', 
	'cn', 
	'objectClass',
	'distinguishedName', 
	'nTSecurityDescriptor', 
	'objectGUID', 
	'instanceType', 
	'whenCreated', 
	'whenChanged', 
	'name', 
	'securityIdentifier', 
	'trustDirection', 
	'trustPartner', 
	'trustPosixOffset', 
	'trustType', 
	'trustAttributes', 
	'flatName', 
	'dSCorePropagationData',
]

class TrustType(enum.Enum):
	DOWNLEVEL = 0x00000001 #): The trusted domain is a Windows domain not running Active Directory.
	UPLEVEL = 0x00000002 #): The trusted domain is a Windows domain running Active Directory.
	MIT = 0x00000003 #): The trusted domain is running a non-Windows, RFC4120-compliant Kerberos distribution. This type of trust is distinguished in that (1) a SID is not required for the TDO, and (2) the default key types include the DES-CBC and DES-CRC encryption types (see [RFC4120] section 8.1).
	DCE = 0x00000004 #): Historical reference; this value is not used in Windows.

class TrustDirection(enum.Enum): #enum.IntFlag << the actual type is intflag, but noone cares
	DISABLED = 0x00000000 #: Absence of any flags. The trust relationship exists but has been disabled.
	INBOUND = 0x00000001 #): The trusted domain trusts the primary domain to perform operations such as name lookups and authentication. If this flag is set, then the trustAuthIncoming attribute is present on this object.
	OUTBOUND = 0x00000002 #: The primary domain trusts the trusted domain to perform operations such as name lookups and authentication. If this flag is set, then the trustAuthOutgoing attribute is present on this object.
	BIDIRECTIONAL = 0x00000003 #: OR'ing of the preceding flags and behaviors representing that both domains trust one another for operations such as name lookups and authentication.

class MSADDomainTrust:
	def __init__(self):
		self.sn = None #str
		self.cn = None #str
		self.distinguishedName = None #dn
		self.objectGUID = None

		self.instanceType = None
		self.whenCreated = None
		self.whenChanged = None
		self.name = None
		
		self.securityIdentifier = None
		self.trustDirection = None
		self.trustPartner = None
		self.trustPosixOffset = None
		self.trustType = None
		self.trustAttributes = None
		self.flatName = None
		self.dSCorePropagationData = None

		
	@staticmethod
	def from_ldap(entry):
		adi = MSADDomainTrust()
		adi.sn = entry['attributes'].get('sn') 
		adi.cn = entry['attributes'].get('cn') 
		adi.distinguishedName = entry['attributes'].get('distinguishedName')
		adi.objectGUID = entry['attributes'].get('objectGUID')
		adi.instanceType = entry['attributes'].get('instanceType')
		adi.whenCreated = entry['attributes'].get('whenCreated')
		adi.whenChanged = entry['attributes'].get('whenChanged')
		adi.name = entry['attributes'].get('name')
		adi.securityIdentifier = entry['attributes'].get('securityIdentifier')
		adi.trustDirection = entry['attributes'].get('trustDirection')
		adi.trustPartner = entry['attributes'].get('trustPartner')
		adi.trustPosixOffset = entry['attributes'].get('trustPosixOffset')
		adi.trustType = entry['attributes'].get('trustType')
		adi.trustAttributes = entry['attributes'].get('trustAttributes')
		adi.flatName = entry['attributes'].get('flatName')
		adi.dSCorePropagationData = entry['attributes'].get('dSCorePropagationData')

		if adi.securityIdentifier is not None:
			adi.securityIdentifier = SID.from_bytes(adi.securityIdentifier)
		if adi.trustType is not None:
			adi.trustType = TrustType(adi.trustType)
		if adi.trustDirection is not None:
			adi.trustDirection = TrustDirection(adi.trustDirection)
		return adi

	def to_dict(self):
		return {
			'sn' : self.sn,
			'cn' : self.cn,
			'distinguishedName' : self.distinguishedName,
			'objectGUID' : self.objectGUID,
			'instanceType' : self.instanceType,
			'whenCreated' : self.whenCreated,
			'whenChanged' : self.whenChanged,
			'name' : self.name,
			'securityIdentifier' : self.securityIdentifier,
			'trustDirection' : self.trustDirection,
			'trustPartner' : self.trustPartner,
			'trustPosixOffset' : self.trustPosixOffset,
			'trustType' : self.trustType,
			'trustAttributes' : self.trustAttributes,
			'flatName' : self.flatName,
			'dSCorePropagationData' : self.dSCorePropagationData,
		}
		
	def get_line(self):
		return '%s %s %s %s' % (self.name, self.trustType, self.trustDirection, self.securityIdentifier)
		
	def __str__(self):
		t = '== MSADDomainTrust ==\r\n'
		t+= 'sn : %s\r\n' % self.sn
		t+= 'cn : %s\r\n' % self.cn
		t+= 'distinguishedName : %s\r\n' % self.distinguishedName
		t+= 'objectGUID : %s\r\n' % self.objectGUID
		t+= 'instanceType : %s\r\n' % self.instanceType
		t+= 'whenCreated : %s\r\n' % self.whenCreated
		t+= 'whenChanged : %s\r\n' % self.whenChanged
		t+= 'name : %s\r\n' % self.name
		t+= 'securityIdentifier : %s\r\n' % str(self.securityIdentifier)
		t+= 'trustDirection : %s\r\n' % self.trustDirection
		t+= 'trustPartner : %s\r\n' % self.trustPartner
		t+= 'trustPosixOffset : %s\r\n' % self.trustPosixOffset
		t+= 'trustType : %s\r\n' % self.trustType
		t+= 'trustAttributes : %s\r\n' % self.trustAttributes
		t+= 'flatName : %s\r\n' % self.flatName
		t+= 'dSCorePropagationData : %s\r\n' % self.dSCorePropagationData
		return t


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import copy
from msldap.authentication.spnego.asn1_structs import *
from asn1crypto.core import OctetString

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-spng/d4f2b41c-5f9e-4e11-98d0-ade76467095d


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-spng/94ccc4f8-d224-495f-8d31-4f58d1af598e
## SPNEGO has been assigned the following object identifier (OID): so.org.dod.internet.security.mechanism.snego (1.3.6.1.5.5.2)

class SPNEGO:
	def __init__(self, mode = 'CLIENT'):
		self.mode = mode
		self.authentication_contexts = {}
		self.original_authentication_contexts = {}
		self.selected_authentication_context = None
		self.selected_mechtype = None
		self.iteration_ctr = 0
	
	def get_copy(self):
		spnego = SPNEGO()
		for ctx_name in self.list_original_conexts():
			spnego.add_auth_context(ctx_name, self.get_original_context(ctx_name))
		return spnego
		
	def list_original_conexts(self):
		"""
		Returns a list of authentication context names available to the SPNEGO authentication.
		"""
		return list(self.original_authentication_contexts.keys())
		
	def get_original_context(self, ctx_name):
		"""
		Returns a copy of the original (not used) authentication context sp[ecified by name.
		You may use this ctx to perform future authentication, as it has the user credentials
		"""
		return copy.deepcopy(self.original_authentication_contexts[ctx_name])
	
	def signing_needed(self):
		return self.selected_authentication_context.signing_needed()
	
	def encryption_needed(self):
		return self.selected_authentication_context.encryption_needed()

	def get_seq_number(self):
		return self.selected_authentication_context.get_seq_number()

	async def unsign(self, data):
		#TODO: IMPLEMENT THIS
		return data

	async def verify(self, data, signature):
		return await self.selected_authentication_context.verify(data, signature)

	async def sign(self, data, message_no, direction='init', reset_cipher = False):
		return await self.selected_authentication_context.sign(data, message_no, direction=direction, reset_cipher = reset_cipher)
		
	async def encrypt(self, data, message_no):
		return await self.selected_authentication_context.encrypt(data, message_no)

	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return await self.selected_authentication_context.decrypt(data, message_no, direction=direction)
		
	def add_auth_context(self, name, ctx):
		"""
		Add an authentication context to the given authentication context name.
		Valid names are:
			'NTLMSSP - Microsoft NTLM Security Support Provider'
			'MS KRB5 - Microsoft Kerberos 5'
			'KRB5 - Kerberos 5'
			'KRB5 - Kerberos 5 - User to User'
			'NEGOEX - SPNEGO Extended Negotiation Security Mechanism'
			
		Context MUST be already set up!
		"""
		self.authentication_contexts[name] = ctx
		self.original_authentication_contexts[name] = copy.deepcopy(ctx)
		
	def select_common_athentication_type(self, mech_types):
		for auth_type_name in self.authentication_contexts:
			if auth_type_name in mech_types:
				return auth_type_name, self.authentication_contexts[auth_type_name]
				
		return None, None
		
	async def process_ctx_authenticate(self, token_data, include_negstate = False, flags = None, seq_number = 0, cb_data = None):
		result, to_continue, err = await self.selected_authentication_context.authenticate(token_data, flags = flags, seq_number = seq_number, cb_data = cb_data)
		if err is not None:
			return None, None, err
		if not result:
			return None, False, None
		response = {}
		if include_negstate == True:
			if to_continue == True:
				response['negState'] = NegState('accept-incomplete')
			else:
				response['negState'] = NegState('accept-completed')
		
		response['responseToken'] = result
		return response, to_continue, None
		
	def get_extra_info(self):
		if hasattr(self.selected_authentication_context, 'get_extra_info'):
			return self.selected_authentication_context.get_extra_info()
		return None
	
	def get_session_key(self):
		return self.selected_authentication_context.get_session_key()

	def get_mechtypes_list(self):
		neghint = {'hintName':'not_defined_in_RFC4178@please_ignore'}
		tokinit = {
			'mechTypes': [MechType(mt) for mt in self.authentication_contexts],
			'negHints': NegHints(neghint),
		}

		negtoken = NegotiationToken({'negTokenInit':NegTokenInit2(tokinit)})
		#spnego = GSS_SPNEGO({'NegotiationToken':negtoken})
		return GSSAPI({'type': GSSType('1.3.6.1.5.5.2'), 'value':negtoken}).dump()
	
	async def authenticate(self, token, flags = None, seq_number = 0, cb_data = None):
		"""
		This function is called (multiple times) during negotiation phase of a protocol to determine hich auth mechanism to be used
		Token is a byte array that is an ASN1 NegotiationToken structure.
		"""
		if self.selected_mechtype is None:
			if token is None:
				#first call to auth, we need to create NegTokenInit2
				#we must list all available auth types, if only one is present then generate initial auth data with it
				
				selected_name = None
				mechtypes = []
				for mechname in self.authentication_contexts:
					selected_name = mechname #only used if there is one!
					mechtypes.append(MechType(mechname))
					
				response = {}
				response['mechTypes'] = MechTypes(mechtypes)

				self.negtypes_store = MechTypes(mechtypes).dump()
					
					
				if len(mechtypes) == 1:
					self.selected_authentication_context = self.authentication_contexts[selected_name]
					self.selected_mechtype = selected_name
					result, to_continue, err = await self.selected_authentication_context.authenticate(None, cb_data=cb_data)
					if err is not None:
						return None, None, err

					if not result:
						return None, False, None
					if str(response['mechTypes'][0]) == '1.2.840.48018.1.2.2':
						response['mechToken'] = KRB5Token(result).to_bytes()
								
						#response['mechToken'] = bytes.fromhex('2a864886f712010202') +  #???????
					else:
						response['mechToken'] = result
						#raise Exception('NTLM as RPC GSSAPI not implemented!')
					
				### First message and ONLY the first message goes out with additional wrapping
				
				negtoken = NegotiationToken({'negTokenInit':NegTokenInit2(response)})
					
					
				#spnego = GSS_SPNEGO({'NegotiationToken':negtoken})
				return GSSAPI({'type': GSSType('1.3.6.1.5.5.2'), 'value':negtoken}).dump(), True, None
					
					
			else:
				#we have already send the NegTokenInit2, but it contained multiple auth types,
				#at this point server is replying which auth type to use
				neg_token_raw = NegotiationToken.load(token)
				
				neg_token = neg_token_raw.native
				
				if not isinstance(neg_token_raw, NegTokenResp):
					raise Exception('Server send init???')
					
					
				self.selected_authentication_context = self.authentication_contexts[neg_token.mechTypes[0]]
				self.selected_mechtype = neg_token['supportedMech']

	
				response, to_continue, err = await self.process_ctx_authenticate(neg_token['responseToken'], flags = flags, seq_number = seq_number, cb_data = cb_data)
				if err is not None:
					return None, None, err
				return NegTokenResp(response).dump(), to_continue, None
					
		else:
			#everything is netotiated, but authentication needs more setps
			neg_token_raw = NegotiationToken.load(token)
			neg_token = neg_token_raw.native
			if neg_token['negState'] == 'accept-completed':
				return None, True, None
			if neg_token['responseToken'] is None:
				# https://tools.ietf.org/html/rfc4178#section-5
				# mechlistmic exchange happening at the end of the authentication
				return None, True, None
				#raise Exception('Should not be here....')
				#print('server mechListMIC: %s' % neg_token['mechListMIC'])
				#res = await self.verify(self.negtypes_store, neg_token['mechListMIC'])
				#print('res %s' % res)
				#print(self.negtypes_store)
				#print(self.negtypes_store.hex())
				#ret = await self.sign(self.negtypes_store, 0)
				#print(ret)
				#print(ret.hex())
				#res = {
				#	'mechListMIC' : ret, 
				#	'negState': NegState('accept-completed')
				#}
				#return NegotiationToken({'negTokenResp':NegTokenResp(res)}).dump(), True, None

			else:
				response, to_continue, err = await self.process_ctx_authenticate(neg_token['responseToken'], flags = flags, seq_number = seq_number, cb_data = cb_data)
				if err is not None:
					return None, None, err
				if not response:
					return None, False, None

				if self.selected_mechtype.startswith('NTLM'):
					response['mechListMIC'] = await self.sign(self.negtypes_store, 0, reset_cipher = True)
					#self.selected_authentication_context.
					#print(response)
				res = NegotiationToken({'negTokenResp':NegTokenResp(response)}).dump()

				return res, to_continue, None
	
def test():
	test_data = bytes.fromhex('a03e303ca00e300c060a2b06010401823702020aa22a04284e544c4d5353500001000000978208e2000000000000000000000000000000000a00d73a0000000f')
	neg_token = NegotiationToken.load(test_data)
	print(neg_token.native)


	test_data_2 = bytes.fromhex('a181ce3081cba0030a0101a10c060a2b06010401823702020aa281b50481b24e544c4d53535000020000000800080038000000158289e2a7314a557bdb11bf000000000000000072007200400000000a0063450000000f540045005300540002000800540045005300540001001200570049004e003200300031003900410044000400120074006500730074002e0063006f007200700003002600570049004e003200300031003900410044002e0074006500730074002e0063006f007200700007000800aec600bfc5fdd40100000000')
	neg_token = NegotiationToken.load(test_data_2)
	print(neg_token.native)

	test_data_3 = bytes.fromhex('a11b3019a0030a0100a3120410010000006b65125a00bb9ab400000000')
	neg_token = NegotiationToken.load(test_data_3)
	print(neg_token.native)

	mt = MechType('NTLMSSP - Microsoft NTLM Security Support Provider')
	print(mt)

	print(MechType.map('1.3.6.1.4.1.311.2.2.10'))
	print(MechType.unmap('1.3.6.1.4.1.311.2.2.10'))

	#spnego_test = SPNEGO()
	#spnego_test.authenticate(test_data_2)
if __name__ == '__main__':
	test()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from msldap.wintypes import *
from msldap.ldap_objects.common import MSLDAP_UAC, vn
from winacl.dtyp.sid import SID

MSADGroup_ATTRS = [ 	
	'cn', 'distinguishedName', 'objectGUID', 'objectSid', 'groupType', 
	'instanceType', 'name', 'member', 'sAMAccountName', 'systemFlags', 
	'whenChanged', 'whenCreated', 'description', 'nTSecurityDescriptor',
	'sAMAccountType',
]


class MSADGroup:
	def __init__(self):
		self.cn = None #str
		self.distinguishedName = None #dn
		self.objectGUID = None
		self.objectSid = None
		self.description = None
		self.groupType = None
		self.instanceType = None
		self.name = None
		self.member = None
		self.nTSecurityDescriptor = None
		self.sAMAccountName = None
		self.sAMAccountType = None
		self.systemFlags = None
		self.whenChanged = None
		self.whenCreated = None
		
	def to_dict(self):
		d = {}
		d['cn'] = self.cn
		d['distinguishedName'] = self.distinguishedName
		d['objectGUID'] = self.objectGUID
		d['objectSid'] = self.objectSid
		d['description'] = self.description
		d['groupType'] = self.groupType
		d['instanceType'] = self.instanceType
		d['name'] = self.name
		d['member'] = self.member
		d['nTSecurityDescriptor'] = self.nTSecurityDescriptor
		d['sAMAccountName'] = self.sAMAccountName
		d['sAMAccountType'] = self.sAMAccountType
		d['systemFlags'] = self.systemFlags
		d['whenChanged'] = self.whenChanged
		d['whenCreated'] = self.whenCreated

		return d	
	
	@staticmethod
	def from_ldap(entry):
		t = MSADGroup()
		t.cn = entry['attributes'].get('cn')
		t.distinguishedName = entry['attributes'].get('distinguishedName')
		t.objectGUID = entry['attributes'].get('objectGUID')
		t.objectSid = entry['attributes'].get('objectSid')
		t.groupType = entry['attributes'].get('groupType')
		t.instanceType = entry['attributes'].get('instanceType')
		t.name = entry['attributes'].get('name')
		t.member = entry['attributes'].get('member')
		t.sAMAccountName = entry['attributes'].get('sAMAccountName')
		t.systemFlags = entry['attributes'].get('systemFlags')
		t.whenChanged = entry['attributes'].get('whenChanged')
		t.whenCreated = entry['attributes'].get('whenCreated')
		
		t.description =  entry['attributes'].get('description')
		if isinstance(t.description, list):
			if len(t.description) == 1:
				t.description = t.description[0]
			else:
				t.description = ', '.join(t.description)
		
		
		#temp = entry['attributes'].get('nTSecurityDescriptor')
		#if temp:
		#	t.nTSecurityDescriptor = SID.from_bytes(temp)
		return t
		

	def __str__(self):
		t = 'MSADGroup\r\n'
		for x in self.__dict__:
			if not isinstance(self.__dict__[x], (list, dict)):
				t += '%s: %s\r\n' % (x, str(self.__dict__[x]))
			else:
				t += '%s: %s\r\n' % (x, self.__dict__[x])
		return t


#
#
# This is just a simple interface to the minikerberos library to support SPNEGO
# 
#
# - Links - 
# 1. See minikerberos library

import datetime

import os
from minikerberos.common import *


from minikerberos.protocol.asn1_structs import AP_REP, EncAPRepPart, EncryptedData, AP_REQ, Ticket
from msldap.authentication.kerberos.gssapi import get_gssapi, KRB5_MECH_INDEP_TOKEN
from msldap.commons.proxy import MSLDAPProxyType
from minikerberos.protocol.structures import ChecksumFlags
from minikerberos.protocol.encryption import Enctype, Key, _enctype_table
from minikerberos.protocol.constants import MESSAGE_TYPE
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.network.aioclientsockssocket import AIOKerberosClientSocksSocket
from msldap import logger

# SMBKerberosCredential

MSLDAP_SOCKS_PROXY_TYPES = [
	MSLDAPProxyType.SOCKS4, 
	MSLDAPProxyType.SOCKS4_SSL, 
	MSLDAPProxyType.SOCKS5, 
	MSLDAPProxyType.SOCKS5_SSL,
	MSLDAPProxyType.WSNET,
]

class MSLDAPKerberos:
	def __init__(self, settings):
		self.settings = settings
		self.signing_preferred = None
		self.encryption_preferred = None
		self.ccred = None
		self.target = None
		self.spn = None
		self.kc = None
		self.flags = None
		self.preferred_etypes = [23,17,18]
		
		self.session_key = None
		self.gssapi = None
		self.iterations = 0
		self.etype = None
		self.seq_number = 0
		self.expected_server_seq_number = None
		self.from_ccache = False
	
		self.setup()
	
	def get_seq_number(self):
		"""
		Returns the initial sequence number. It is 0 by default, but can be adjusted during authentication, 
		by passing the 'seq_number' parameter in the 'authenticate' function
		"""
		return self.seq_number
	
	def signing_needed(self):
		"""
		Checks if integrity protection was negotiated
		"""
		return ChecksumFlags.GSS_C_INTEG_FLAG in self.flags
	
	def encryption_needed(self):
		"""
		Checks if confidentiality flag was negotiated
		"""
		return ChecksumFlags.GSS_C_CONF_FLAG in self.flags
				
	async def sign(self, data, message_no, direction = 'init'):
		"""
		Signs a message. 
		"""
		return self.gssapi.GSS_GetMIC(data, message_no, direction = direction)	
		
	async def encrypt(self, data, message_no):
		"""
		Encrypts a message. 
		"""

		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init'):
		"""
		Decrypts message. Also performs integrity checking.
		"""

		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction)
		
	def setup(self):
		self.ccred = self.settings.ccred
		self.spn = self.settings.spn
		self.target = self.settings.target
		if self.settings.enctypes is not None:
			self.preferred_etypes = self.settings.enctypes
		
		self.flags = ChecksumFlags.GSS_C_MUTUAL_FLAG
		if self.settings.encrypt is True:
			self.flags = \
				ChecksumFlags.GSS_C_CONF_FLAG |\
				ChecksumFlags.GSS_C_INTEG_FLAG |\
				ChecksumFlags.GSS_C_REPLAY_FLAG |\
				ChecksumFlags.GSS_C_SEQUENCE_FLAG
		
	
	def get_session_key(self):
		return self.session_key.contents, None


	async def setup_kc(self):
		try:
			# sockst/wsnet proxying is handled by the minikerberos&asysocks modules
			if self.target.proxy is None or self.target.proxy.type in MSLDAP_SOCKS_PROXY_TYPES:
				self.kc = AIOKerberosClient(self.ccred, self.target)

			elif self.target.proxy.type in [MSLDAPProxyType.MULTIPLEXOR, MSLDAPProxyType.MULTIPLEXOR_SSL]:
				from msldap.network.multiplexor import MultiplexorProxyConnection
				mpc = MultiplexorProxyConnection(self.target)
				socks_proxy = await mpc.connect(is_kerberos = True)

				self.kc = AIOKerberosClient(self.ccred, socks_proxy)

			else:
				raise Exception('Unknown proxy type %s' % self.target.proxy.type)

			return None, None
		except Exception as e:
			return None, e
	
	async def authenticate(self, authData, flags = None, seq_number = 0, cb_data = None):
		"""
		This function is called (multiple times depending on the flags) to perform authentication. 
		"""
		try:
			if self.kc is None:
				_, err = await self.setup_kc()
				if err is not None:
					return None, None, err

			if self.iterations == 0:
				self.seq_number = 0
				self.iterations += 1
				
				try:
					#check TGS first, maybe ccache already has what we need
					for target in self.ccred.ccache.list_targets():
						# just printing this to debug...
						logger.debug('CCACHE SPN record: %s' % target)
					tgs, encpart, self.session_key = await self.kc.get_TGS(self.spn)
					
					self.from_ccache = True
				except:
					tgt = await self.kc.get_TGT(override_etype = self.preferred_etypes)
					tgs, encpart, self.session_key = await self.kc.get_TGS(self.spn)#, override_etype = self.preferred_etypes)

				#self.expected_server_seq_number = encpart.get('nonce', seq_number)
				
				ap_opts = []
				if ChecksumFlags.GSS_C_MUTUAL_FLAG in self.flags or ChecksumFlags.GSS_C_DCE_STYLE in self.flags:
					if ChecksumFlags.GSS_C_MUTUAL_FLAG in self.flags:
						ap_opts.append('mutual-required')
					if self.from_ccache is False:
						apreq = self.kc.construct_apreq(tgs, encpart, self.session_key, flags = self.flags, seq_number = self.seq_number, ap_opts=ap_opts, cb_data = cb_data)
					else:
						apreq = self.kc.construct_apreq_from_ticket(Ticket(tgs['ticket']).dump(), self.session_key, tgs['crealm'], tgs['cname']['name-string'][0], flags = self.flags, seq_number = self.seq_number, ap_opts = ap_opts, cb_data = cb_data)
					return apreq, True, None
				
				else:
					#no mutual or dce auth will take one step only
					if self.from_ccache is False:
						apreq = self.kc.construct_apreq(tgs, encpart, self.session_key, flags = self.flags, seq_number = self.seq_number, ap_opts=[], cb_data = cb_data)
					else:
						apreq = self.kc.construct_apreq_from_ticket(Ticket(tgs['ticket']).dump(), self.session_key, tgs['crealm'], tgs['cname']['name-string'][0], flags = self.flags, seq_number = self.seq_number, ap_opts = ap_opts, cb_data = cb_data)
					
					
					self.gssapi = get_gssapi(self.session_key)
					return apreq, False, None

			else:
				self.iterations += 1
				if ChecksumFlags.GSS_C_DCE_STYLE in self.flags:
					# adata = authData[16:]
					# if ChecksumFlags.GSS_C_DCE_STYLE in self.flags:
					#	adata = authData
					raise Exception('DCE auth Not implemented!')
				
				# at this point we are dealing with mutual authentication
				# This means that the server sent back an AP-rep wrapped in a token
				# The APREP contains a new session key we'd need to update and a seq-number 
				# that is expected the server will use for future communication.
				# For mutual auth we dont need to reply anything after this step, 
				# but for DCE auth a reply is expected. TODO

				# converting the token to aprep
				token = KRB5_MECH_INDEP_TOKEN.from_bytes(authData)
				if token.data[:2] != b'\x02\x00':
					raise Exception('Unexpected token type! %s' % token.data[:2].hex() )
				aprep = AP_REP.load(token.data[2:]).native
				
				# decrypting aprep
				cipher = _enctype_table[int(aprep['enc-part']['etype'])]()
				cipher_text = aprep['enc-part']['cipher']
				temp = cipher.decrypt(self.session_key, 12, cipher_text)
				enc_part = EncAPRepPart.load(temp).native

				#updating session key, gssapi
				self.session_key = Key(int(enc_part['subkey']['keytype']), enc_part['subkey']['keyvalue'])
				#self.seq_number = enc_part.get('seq-number', 0)
				self.gssapi = get_gssapi(self.session_key)

				return b'', False, None		
		
		except Exception as e:
			return None, None, e



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import enum

import platform
import ssl


class LDAPProtocol(enum.Enum):
	TCP = 'TCP'
	UDP = 'UDP'
	SSL = 'SSL'


class MSLDAPTarget:
	"""
	Describes the connection to the server.
	
	:param host: IP address or hostname of the server
	:type host: str
	:param port: port of the LDAP service running on the server
	:type port: int
	:param proto: Connection protocol to be used
	:type proto: :class:`LDAPProtocol`
	:param tree: The tree to connect to
	:type tree: str
	:param proxy: specifies what kind of proxy to be used
	:type proxy: :class:`MSLDAPProxy`
	:param timeout: connection timeout in seconds
	:type timeout: int
	:param ldap_query_page_size: Maximum number of elements to fetch in each paged_query call.
	:type ldap_query_page_size: int
	:param ldap_query_ratelimit: rate limit of paged queries. This will cause a sleep (in seconds) between fetching of each page of the query
	:type ldap_query_ratelimit: float
	:param dc_ip: Ip address of the kerberos server (if kerberos is used)
	:type dc_ip: str
	"""
	def __init__(self, host, port = 389, proto = LDAPProtocol.TCP, tree = None, proxy = None, timeout = 10, ldap_query_page_size = 1000, ldap_query_ratelimit = 0, dc_ip:str = None):
		self.proto = proto
		self.host = host
		self.tree = tree
		self.port = port
		self.proxy = proxy
		self.timeout = timeout
		self.dc_ip = dc_ip
		self.serverip = None
		self.domain = None
		self.sslctx = None
		self.ldap_query_page_size = ldap_query_page_size
		self.ldap_query_ratelimit = ldap_query_ratelimit

	def get_ssl_context(self):
		if self.proto == LDAPProtocol.SSL:
			if self.sslctx is None:
				# TODO ssl verification :)
				self.sslctx = ssl._create_unverified_context()
				#self.sslctx.verify = False
			return self.sslctx
		return None

	def to_target_string(self):
		return 'ldap/%s@%s' % (self.host,self.domain)  #ldap/WIN2019AD.test.corp @ TEST.CORP

	def get_host(self):
		return '%s://%s:%s' % (self.proto, self.host, self.port)

	def is_ssl(self):
		return self.proto == LDAPProtocol.SSL
	
	def __str__(self):
		t = '==== MSLDAPTarget ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t